# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from dataclasses import dataclass
from math import prod

from .backend import ArrayLike, get_index_dtype, size, shape
from .trainbase import TrainBase
from .utils import namespace_of_trains, symbol_generator
from .exactaddition import ExactAddition
from .evaluationexpression import EvaluationExpression
from .exactaddition import ExactAddition
from .full import full

@dataclass
class MinMaxResult[T: ArrayLike]:
    """
    Result of the min_max algorithm. Contains the values and indices of the minimum and maximum
    values found in a tensor train.
    """

    #: Indices of the minimum value
    min_idxs: T
    #: Minimum value
    min_val: float
    #: Indices of the maximum value
    max_idxs: T
    #: Maximum value
    max_val: float

def min_max[T: ArrayLike](train: TrainBase[T], num: int) -> MinMaxResult[T]:
    xp = namespace_of_trains(train)

    sgen = symbol_generator()
    dim_chars = "".join(next(sgen) for _ in range(len(train.shape.dims)))
    ev_expr = EvaluationExpression(f"{dim_chars}->{dim_chars}", train.shape)

    right_idxs = right_sweep(train, num)
    right_idxs = xp.reshape(right_idxs, (len(train.shape.dims), 1))
    right_val = float(ev_expr(right_idxs, train)[0])

    left_idxs = left_sweep(train, num)
    left_idxs = xp.reshape(left_idxs, (len(train.shape.dims), 1))
    left_val = float(ev_expr(left_idxs, train)[0])

    if abs(left_val) < abs(right_val):
        idxs, val = right_idxs, right_val
    else:
        idxs, val = left_idxs, left_val

    tmp = full(xp, train.shape, -val)
    mod_train = ExactAddition()(train, tmp)

    mod_right_idxs = right_sweep(mod_train, num)
    mod_right_idxs = xp.reshape(mod_right_idxs, (len(train.shape.dims), 1))
    mod_right_val = float(ev_expr(mod_right_idxs, train)[0])

    mod_left_idxs = left_sweep(mod_train, num)
    mod_left_idxs = xp.reshape(mod_left_idxs, (len(train.shape.dims), 1))
    mod_left_val = float(ev_expr(mod_left_idxs, train)[0])

    if abs(mod_left_val) < abs(mod_right_val):
        mod_idxs, mod_val = mod_right_idxs, mod_right_val
    else:
        mod_idxs, mod_val = mod_left_idxs, mod_left_val

    if val < mod_val:
        return MinMaxResult(idxs, val, mod_idxs, mod_val)
    return MinMaxResult(mod_idxs, mod_val, idxs, val)

def right_sweep[T: ArrayLike](train: TrainBase[T], num: int) -> T:
    xp = namespace_of_trains(train)
    tshape = train.shape
    ndims = len(tshape.dims)
    max_middle = max(prod(tshape.middle(i)) for i in range(len(tshape)))

    cur_idxs = xp.zeros((ndims, max_middle*num),
                        device=train.device, dtype=get_index_dtype(xp))
    tmp_idxs = xp.zeros((ndims, max_middle*num),
                        device=train.device, dtype=get_index_dtype(xp))
    dim_map = {dim.idf: i for i, dim in enumerate(train.shape.dims)}

    cur_num = 1
    mat = xp.ones((1, 1), device=train.device, dtype=train.dtype)
    train.normalize(0)
    for i, tn in enumerate(train.data):
        middle = prod(tshape.middle(i))
        mat = xp.tensordot(mat, tn, axes=([1],[0]))
        mat = xp.reshape(mat, (prod(shape(mat)[:-1]), mat.shape[-1]))

        for j in range(middle):
            cut = (slice(None), slice(j, cur_num*middle, middle))
            tmp_idxs[cut] = cur_idxs[:,:cur_num]
        for digit in tshape.digits[i]:
            idx = dim_map[digit.idf]
            for j in range(digit.base):
                cut = (idx, slice(j, cur_num*middle, digit.base))
                tmp_idxs[cut] += digit.factor * j

        norms = xp.sqrt(xp.sum(xp.pow(mat, 2), axis=1))
        cur_num = min(num, size(norms))
        norm_idxs = xp.argsort(norms, descending=True)[:cur_num]# topk behaviour
        cur_idxs[:,:size(norm_idxs)] = xp.take(tmp_idxs, norm_idxs, axis=1)
        mat = xp.take(mat, norm_idxs, axis=0)
    return cur_idxs[:,0]

def left_sweep[T: ArrayLike](train: TrainBase[T], num: int) -> T:
    xp = namespace_of_trains(train)
    tshape = train.shape
    ndims = len(tshape.dims)
    max_middle = max(prod(tshape.middle(i)) for i in range(len(tshape)))

    cur_idxs = xp.zeros((ndims, max_middle*num),
                        device=train.device, dtype=get_index_dtype(xp))
    tmp_idxs = xp.zeros((ndims, max_middle*num),
                        device=train.device, dtype=get_index_dtype(xp))
    dim_map = {dim.idf: i for i, dim in enumerate(train.shape.dims)}

    cur_num = 1
    mat = xp.ones((1, 1), device=train.device, dtype=train.dtype)
    train.normalize(-1)
    for i, tn in zip(reversed(range(len(tshape))), reversed(train.data)):
        middle = prod(tshape.middle(i))
        idx = len(tn.shape) - 1
        mat = xp.tensordot(tn, mat, axes=([idx],[0]))
        mat = xp.reshape(mat, (mat.shape[0], prod(shape(mat)[1:])))

        for j in range(middle):
            cut = (slice(None), slice(j*cur_num, (j+1)*cur_num))
            tmp_idxs[cut] = cur_idxs[:,:cur_num]
        for digit in tshape.digits[i]:
            idx = dim_map[digit.idf]
            off = (cur_num*middle)//digit.base
            for j in range(digit.base):
                cut = (idx, slice(j*off, (j+1)*off))
                tmp_idxs[cut] += digit.factor * j
        norms = xp.sqrt(xp.sum(xp.pow(mat, 2), axis=0))
        cur_num = min(num, size(norms))
        norm_idxs = xp.argsort(norms, descending=True)[:cur_num]# topk behaviour
        cur_idxs[:,:size(norm_idxs)] = xp.take(tmp_idxs, norm_idxs, axis=1)
        mat = xp.take(mat, norm_idxs, axis=1)
    return cur_idxs[:,0]



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Literal, Optional, Sequence, overload
from types import NoneType
from copy import deepcopy
from math import prod

from .digit import Digit, Digits, digits_similar
from .dimension import Dimension, dimensions_similar

class TrainShape:
    """
    Class for representing the structure of a quantics tensor train.
    The shape consists of one or multiple dimensions, which are each
    represented as a sequence of digits. The digits are grouped into
    cores that are connected linearly. 
    """

    _dims: Sequence[Dimension]
    _ranks: list[int]
    _digits: list[Digits]

    @property
    def dims(self) -> Sequence[Dimension]:
        """Dimensions of the shape. Cannot be set."""
        return self._dims

    @property
    def digits(self) -> Sequence[Digits]:
        """Digits that are grouped to cores. Cannot be set."""
        return self._digits

    @property
    def ranks(self) -> Sequence[int]:
        """
        Ranks between the cores. The length of ranks is one less than the number of cores.
        Setting the ranks to None will set the ranks to the maximum possible values, for
        representing a tensor exactly. The ranks can also be set to an integer, which will
        set all ranks to the minimum of the given integer and the maximum possible value.
        Otherwise, the ranks can be set individually by providing a sequence of integers.
        """
        return self._ranks
    @ranks.setter
    def ranks(self, ranks: NoneType | int | Sequence[int] = None, /) -> None:
        if ranks is None:
            self._ranks = self._exact_ranks()
        elif isinstance(ranks, int):
            exact_ranks = self._exact_ranks()
            self._ranks = [min(rank, ranks) for rank in exact_ranks]
        else:
            if len(ranks) != len(self._ranks):
                raise ValueError(f"Length of ranks {len(ranks)} does not match length of shape {len(self._ranks)}.")
            self._ranks = list(ranks)

    def __init__(
            self,
            dims: Dimension | Sequence[Dimension],
            cores: Sequence[Digits],
            ranks: NoneType | int | Sequence[int] = 1) -> None:
        if isinstance(dims, Dimension):
            dims = [dims]
        self._cores_match_dims(dims, cores)
        self._distinct_cores(cores)

        self._dims = list(deepcopy(dim) for dim in dims)
        self._digits = list(deepcopy(core) for core in cores)
        self._ranks = [0] * (len(cores)-1)
        self.ranks = ranks

    # ------------------------------------------------------------------------
    # shape getter

    def left_rank(self, idx: int) -> int:
        """Get the left rank of the core at index idx."""
        idx = transform_index(idx, len(self))
        if idx == 0:
            return 1
        return self._ranks[idx-1]

    def right_rank(self, idx: int) -> int:
        """Get the right rank of the core at index idx."""
        idx = transform_index(idx, len(self))
        if idx == len(self)-1:
            return 1
        return self._ranks[idx]

    def middle(self, idx: int) -> Sequence[int]:
        """Get the bases of the digits in the core at index idx."""
        idx = transform_index(idx, len(self))
        return [d.base for d in self._digits[idx]]

    def reverse(self) -> TrainShape:
        """return a new TrainShape with reversed order of cores and digits."""
        cores = [list(reversed(digits)) for digits in reversed(self._digits)]
        ranks = list(reversed(self._ranks))
        return TrainShape(self._dims, cores, ranks)

    # ------------------------------------------------------------------------
    # magic stuff

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TrainShape):
            return False
        flag = len(self.dims) == len(other.dims)\
           and len(self) == len(other)\
           and all(dim1==dim2 for dim1, dim2 in zip(self.dims, other.dims))
        if not flag:
            return False
        idf_map = {dim1.idf: dim2.idf for dim1, dim2 in zip(self.dims, other.dims)}
        return all(_similar(dgts1, dgts2, idf_map) for dgts1, dgts2 in zip(self.digits, other.digits))

    def __len__(self) -> int:
        return len(self._digits)

    def __repr__(self) -> str:
        res = ""
        for i, digits in enumerate(self._digits[:-1]):
            res += f"[{','.join(str(d.base) for d in digits)}]"
            res += f"--{self._ranks[i]}--"
        res += f"[{','.join(str(d.base) for d in self._digits[-1])}]"
        return res

    # ------------------------------------------------------------------------
    # helper

    def _exact_ranks(self) -> list[int]:
        if len(self._digits) == 1:
            return []
        tmp1 = [prod(self.middle(0))]
        for i in range(1, len(self)-1):
            tmp1.append(tmp1[-1]*prod(self.middle(i)))
        tmp2 = [prod(self.middle(-1))]
        for i in reversed(range(1, len(self)-1)):
            tmp2.append(tmp2[-1]*prod(self.middle(i)))
        return [min(t1, t2) for t1, t2 in zip(tmp1, reversed(tmp2))]

    def _cores_match_dims(self, dims: Sequence[Dimension], cores: Sequence[Digits]):
        core_digits = set(digit for core in cores for digit in core)
        dim_digits = set(digit for dim in dims for digit in dim)
        if core_digits != dim_digits:
            raise ValueError("Digits in cores and dimensions do not match")

    def _distinct_cores(self, cores: Sequence[Digits]) -> None:
        digits = [digit for core in cores for digit in core]
        ref = set(digits)
        if len(digits) != len(ref):
            raise ValueError("Cores must have distinct digits")

def transform_index(idx: int, length: int) -> int:
    if idx < 0:
        idx += length
    if idx < 0 or idx >= length:
        raise IndexError(f"Index {idx} out of range for sequence of length {length}.")
    return idx

def change_dims(shape: TrainShape, dims: Sequence[Dimension]) -> TrainShape:
    """Change the dimensions of a TrainShape."""
    if len(dims) != len(shape.dims):
        raise ValueError("New dimensions must have the same length as the old ones.")
    if not all(dimensions_similar(old_dim, new_dim) for old_dim, new_dim in zip(shape.dims, dims)):
        raise ValueError("New dimensions must be similar to the old ones.")
    idf_map = {old_dim.idf: new_dim.idf for old_dim, new_dim in zip(shape.dims, dims)}
    cores = []
    for i in range(len(shape)):
        digits = [Digit(idf_map[digit.idf], digit.idx, digit.base, digit.factor) for digit in shape.digits[i]]
        cores.append(digits)
    return TrainShape(dims, cores)

@overload
def trainshape(*dims: Dimension, digits: Sequence[Digits]) -> TrainShape: ...
@overload
def trainshape(*dims: int | Dimension, mode: Literal["full", "block", "interleaved"] = "block") -> TrainShape: ...
# implementation
def trainshape(
        *dims: int | Dimension,
        mode: Literal["full", "block", "interleaved"] = "block",
        digits: Optional[Sequence[Digits]] = None
        ) -> TrainShape:
    dims = tuple([Dimension(dim) if isinstance(dim, int) else dim for dim in dims])
    if digits is not None:
        return TrainShape(dims, digits)
    if mode == "block":
        return TrainShape(dims, [(d,) for dim in dims for d in dim])
    elif mode == "interleaved":
        length = max(len(dim) for dim in dims)
        digits = []
        for i in range(length):
            tmp = []
            for dim in dims:
                if i < len(dim):
                    tmp.append(dim[i])
            digits.append(tmp)
        return TrainShape(dims, digits)
    elif mode == "full":
        digits = []
        for dim in dims:
            digits.extend(dim)
        return TrainShape(dims, [digits])
    raise ValueError("Invalid mode. Choose 'block' or 'interleaved'.")

def _similar(digit1: Digits, digit2: Digits, idf_map: dict[int, int]) -> bool:
    if len(digit1) != len(digit2):
        return False
    return all(digits_similar(d1, d2) and d2.idf == idf_map[d1.idf] for d1, d2 in zip(digit1, digit2))



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from types import NoneType
from dataclasses import dataclass
from copy import deepcopy
from collections import deque

from .backend import ArrayLike, namespace_of_arrays, size
from .digit import Digits
from .normalization import Normalization
from .matrixdecomposition import MatrixDecompositionResult
from .tensordecomposition import TensorDecomposition

@dataclass(kw_only=True)
class CutResult[T: ArrayLike, S: MatrixDecompositionResult]:
    shape: Digits
    norm: Normalization
    data: T
    decomp_result: NoneType | S = None

class SuperCore[T: ArrayLike]:

    _digit_queue: deque[Digits]
    _data: NoneType | T
    norm: Normalization

    @property
    def data(self) -> T:
        if self._data is None:
            raise ValueError("Supercore has no data yet.")
        return self._data
    @data.setter
    def data(self, value: T) -> None:
        if self._data is None:
            raise ValueError("Cannot set data without adding cores.")
        if value.shape[1:-1] != self._data.shape[1:-1]:
            raise ValueError(f"New data shape does not match existing shape.")
        if size(value) == 0:
            raise ValueError("Cannot set data with a size of zero.")
        self._data = value
        self.norm = Normalization.NONE

    @property
    def shapes(self) -> Sequence[Digits]:
        return self._digit_queue

    #-------------------------------------------------------------------------
    #constructor & methods

    def __init__(self) -> None:
        self._digit_queue = deque()
        self._norm = Normalization.NONE
        self._data = None

    def set_data(self, cores: Sequence[Digits], data: T, norm: Normalization = Normalization.NONE) -> None:
        if len(cores) == 0:
            raise ValueError("Cannot set data with empty core sequence.")
        if size(data) == 0:
            raise ValueError("Cannot set data with a size of zero.")
        core_middle = []
        for core in cores:
            core_middle.extend([d.base for d in core])
        if tuple(data.shape[1:-1]) != tuple(core_middle):
            raise ValueError(f"Shape of cores and data shape do not match: ")
        self._digit_queue = deque(deepcopy(core) for core in cores)
        self._data = data
        self._norm = norm

    def add_right(self, digits: Digits, data: T) -> None:
        if self._data is None:
            self._add_empty(digits, data)
            return
        if tuple(d.base for d in digits) != tuple(data.shape[1:-1]):
            raise ValueError(f"Core shape and data shape do not match: ")
        if self._data.shape[-1] != data.shape[0]:
            raise ValueError(f"Added core has invalid left rank")

        xp = namespace_of_arrays(data)
        self._data = xp.tensordot(self._data, data, axes=([-1], [0]))
        self._digit_queue.append(deepcopy(digits))
        self._norm = Normalization.NONE
        
    def add_left(self, digits: Digits, data: T) -> None:
        if self._data is None:
            self._add_empty(digits, data)
            return
        if tuple(d.base for d in digits) != tuple(data.shape[1:-1]):
            raise ValueError(f"Core shape and data shape do not match: ")
        if self._data.shape[0] != data.shape[-1]:
            raise ValueError(f"Added core has invalid right rank")

        xp = namespace_of_arrays(data)
        self._data = xp.tensordot(data, self._data, axes=([-1], [0]))
        self._digit_queue.appendleft(deepcopy(digits))
        self._norm = Normalization.NONE

    def _add_empty(self, digits: Digits, data: T) -> None:
        self._digit_queue.append(deepcopy(digits))
        self._data = data

    def cut_right[S: MatrixDecompositionResult](
            self,
            decomp: TensorDecomposition[T, S]
            ) -> CutResult[T, S]:
        if self._data is None or len(self._digit_queue) == 0:
            raise ValueError("Supercore has no data to cut.")
        if len(self._digit_queue) == 1:
            return self._clear(decomp)

        digits = self._digit_queue.pop()
        split = len(digits)+1
        res = decomp.right(self._data, -split)
        self._data, data = res.left, res.right
        self._norm = Normalization.NONE
        return CutResult(shape=digits, norm=Normalization.RIGHT,
                         data=data, decomp_result=res)

    def cut_left[S: MatrixDecompositionResult](
            self,
            decomp: TensorDecomposition[T, S]
            ) -> CutResult[T, S]:
        if self._data is None or len(self._digit_queue) == 0:
            raise ValueError("Supercore has no data to cut.")
        if len(self._digit_queue) == 1:
            return self._clear(decomp)

        digits = self._digit_queue.popleft()
        split = len(digits)+1
        res = decomp.left(self._data, split)
        data, self._data = res.left, res.right
        self._norm = Normalization.NONE
        return CutResult(shape=digits, norm=Normalization.LEFT,
                         data=data, decomp_result=res)

    def _clear[S: MatrixDecompositionResult](
            self,
            _: TensorDecomposition[T, S]
            ) -> CutResult[T, S]:
        if self._data is None:
            raise ValueError("Supercore has no data to clear.")
        if len(self._digit_queue) != 1:
            raise Exception("Supercore has more than one core, cannot clear.")
        core = self._digit_queue.pop()
        #core.left = self._data.shape[0]
        #core.right = self._data.shape[-1]
        data = self._data
        norm = self._norm
        self._data = None
        self._norm = Normalization.NONE
        #return core, data, norm
        return CutResult(shape=core, norm=norm, data=data)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from types import EllipsisType
from typing import Any, Protocol, Sequence, Self, TypeVar
from collections.abc import Buffer

ArrayLike = Any
DType = Any
Device = Any
#ArrayNamespace = Any

NestedSequence = Sequence[Any]

"""
class Device(Protocol):
    pass

class DType(Protocol):
    def __eq__(self, other: Any) -> bool: ...

class ArrayLike(Protocol):

    Slice = int | slice | EllipsisType | None | Sequence[int | slice | EllipsisType | Self | None] | Self
    Shape = tuple[int | None, ...]

    @property
    def device(self) -> Any: ...

    @property
    def dtype(self) -> Any: ...

    @property
    def shape(self) -> Shape: ...

    @property
    def ndim(self) -> int: ...

    @property
    def mT(self) -> Self: ...

    @property
    def T(self) -> Self: ...

    def __add__(self, other: int | float | complex | Self, /) -> Self: ...
    def __abs__(self) -> Self: ...
    def __and__(self, other: int | bool | Self, /) -> Self: ...
    def __array_namespace__(self,  api_version: str | None = None) -> ArrayNamespace: ...
    def __bool__(self) -> bool: ...
    def __complex__(self) -> complex: ...
    def __dlpack__(self) -> Any: ...
    def __dlpack_device__(self) -> Any: ...
    def __eq__(self, other: int | float | complex | bool | Self, /) -> Self: ... # type: ignore
    def __float__(self) -> float: ...
    def __floordiv__(self, other: int | float | Self, /) -> Self: ...
    def __ge__(self, other: int | float | Self, /) -> Self: ...
    def __getitem__(self, key: Slice, /) -> Self: ...
    def __gt__(self, other: int | float | Self, /) -> Self: ...
    def __index__(self, *args, **kwargs) -> int: ...
    def __int__(self) -> int: ...
    def __invert__(self) -> Self: ...
    def __le__(self, other: int | float | Self, /) -> Self: ...
    def __lshift__(self, other: int | Self, /) -> Self: ...
    def __lt__(self, other: int | float | Self, /) -> Self: ...
    def __matmul__(self, other: Self, /) -> Self: ...
    def __mod__(self, other: int | float | Self) -> Self: ...
    def __mul__(self, other: int | float | complex | Self, /) -> Self: ...
    def __ne__(self, other: int | float | complex | bool | Self, /) -> Self: ... # type: ignore
    def __neg__(self) -> Self: ...
    def __or__(self, other: int | bool | Self) -> Self: ...
    def __pos__(self) -> Self: ...
    def __pow__(self, other: int | float | complex | Self, /) -> Self: ...
    def __rshift__(self, other: int | Self, /) -> Self: ...
    def __setitem__(self, key: Slice, value: int | float | complex | bool | Self, /) -> None: ...
    def __sub__(self, other: int | float | complex | Self, /) -> Self: ...
    def __truediv__(self, other: int | float | complex | Self, /) -> Self: ...
    def __xor__(self, other: int | bool | Self) -> Self: ...

ArrayBound = TypeVar("ArrayBound", bound=ArrayLike)
"""

class Finfo(Protocol):
    @property
    def bits(self) -> int: ...
    @property
    def eps(self) -> float: ...
    @property
    def max(self) -> float: ...
    @property
    def min(self) -> float: ...
    @property
    def smallest_normal(self) -> float: ...
    @property
    def dtype(self) -> DType: ...

class Iinfo(Protocol):
    @property
    def bits(self) -> int: ...
    @property
    def max(self) -> int: ...
    @property
    def min(self) -> int: ...
    @property
    def dtype(self) -> DType: ...

class ArrayNamespaceInfo(Protocol):
    def capabilities(self) -> dict[str, Any]: ...
    def default_device(self) -> Device: ...
    def default_dtype(self, *args, **kwargs) -> DType: ...
    def devices(self) -> list[Device]: ...
    def dtypes(self, *args, **kwargs) -> dict[str, DType]: ...

class LinearAlgebraExtensions[T: ArrayLike](Protocol):
    def cholesky(self, *args, **kwargs) -> T: ...
    def cross(self, *args, **kwargs) -> T: ...
    def det(self, *args, **kwargs) -> T: ...
    def diagonal(self, *args, **kwargs) -> T: ...
    def eigh(self, *args, **kwargs) -> tuple[T, T]: ...
    def eigvalsh(self, *args, **kwargs) -> T: ...
    def inv(self, *args, **kwargs) -> T: ...
    def matmul(self, *args, **kwargs) -> T: ...
    def matrix_norm(self, *args, **kwargs) -> T: ...
    def matrix_power(self, *args, **kwargs) -> T: ...
    def matrix_rank(self, *args, **kwargs) -> T: ...
    def matrix_transpose(self, *args, **kwargs) -> T: ...
    def outer(self, *args, **kwargs) -> T: ...
    def pinv(self, *args, **kwargs) -> T: ...
    def qr(self, *args, **kwargs) -> tuple[T, T]: ...
    def slogdet(self, *args, **kwargs) -> tuple[T, T]: ...
    def solve(self, *args, **kwargs) -> T: ...
    def svd(self, *args, **kwargs) -> tuple[T, T, T]: ...
    def svdvals(self, *args, **kwargs) -> T: ...
    def tensordot(self, *args, **kwargs) -> T: ...
    def trace(self, *args, **kwargs) -> T: ...
    def vecdot(self, *args, **kwargs) -> T: ...
    def vector_norm(self, *args, **kwargs) -> T: ...


class ArrayNamespace[T: ArrayLike](Protocol):

    # Constants
    @property
    def e(self) -> float: ...
    @property
    def pi(self) -> float: ...
    @property
    def inf(self) -> float: ...
    @property
    def nan(self) -> float: ...
    @property
    def newaxis(self) -> None: ...

    @property
    def __array_api_version__(self) -> str: ...

    @property
    def linalg(self) -> LinearAlgebraExtensions: ...

    # Creation functions
    def arange(
            self,
            start: int | float, /,
            stop: int | float | None = None,
            step: int | float = 1, *,
            dtype: DType | None = None,
            device: Device | None = None
            ) -> T: ...
    def asarray(
            self,
            obj: T | bool | int | float | complex | NestedSequence | Buffer, /, *,
            dtype: DType | None = None,
            device: Device | None = None,
            copy: bool | None = None
            ) -> T: ...
    def empty(
            self,
            shape: int | tuple[int, ...], *,
            dtype: DType | None = None,
            device: Device | None = None
            ) -> T: ...
    def empty_like(
            self,
            x: T, /, *,
            dtype: DType | None = None,
            device: Device | None = None
            ) -> T: ...
    def eye(self, *args, **kwargs) -> T: ...
    def from_dlpack(self, *args, **kwargs) -> T: ...
    def full(self, *args, **kwargs) -> T: ...
    def full_like(self, *args, **kwargs) -> T: ...
    def linspace(self, *args, **kwargs) -> T: ...
    def meshgrid(self, *args, **kwargs) -> list[T]: ...
    def ones(self, *args, **kwargs) -> T: ...
    def ones_like(self, *args, **kwargs) -> T: ...
    def tril(self, *args, **kwargs) -> T: ...
    def triu(self, *args, **kwargs) -> T: ...
    def zeros(self, *args, **kwargs) -> T: ...
    def zeros_like(self, *args, **kwargs) -> T: ...

    # Data Type Functions
    def astype(self, *args, **kwargs) -> T: ...
    def can_cast(self, *args, **kwargs) -> bool: ...
    def finfo(self, *args, **kwargs) -> Finfo: ...
    def iinfo(self, *args, **kwargs) -> Iinfo: ...
    def isdtype(self, *args, **kwargs) -> bool: ...
    def result_type(self, *args, **kwargs) -> DType: ...

    # Element-wise functions
    def abs(self, *args, **kwargs) -> T: ...
    def acos(self, *args, **kwargs) -> T: ...
    def acosh(self, *args, **kwargs) -> T: ...
    def add(self, *args, **kwargs) -> T: ...
    def asin(self, *args, **kwargs) -> T: ...
    def asinh(self, *args, **kwargs) -> T: ...
    def atan(self, *args, **kwargs) -> T: ...
    def atan2(self, *args, **kwargs) -> T: ...
    def atanh(self, *args, **kwargs) -> T: ...
    def bitwise_and(self, *args, **kwargs) -> T: ...
    def bitwise_left_shift(self, *args, **kwargs) -> T: ...
    def bitwise_invert(self, *args, **kwargs) -> T: ...
    def bitwise_or(self, *args, **kwargs) -> T: ...
    def bitwise_right_shift(self, *args, **kwargs) -> T: ...
    def bitwise_xor(self, *args, **kwargs) -> T: ...
    def ceil(self, *args, **kwargs) -> T: ...
    def clip(self, *args, **kwargs) -> T: ...
    def conj(self, *args, **kwargs) -> T: ...
    def copysign(self, *args, **kwargs) -> T: ...
    def cos(self, *args, **kwargs) -> T: ...
    def cosh(self, *args, **kwargs) -> T: ...
    def divide(self, *args, **kwargs) -> T: ...
    def equal(self, *args, **kwargs) -> T: ...
    def exp(self, *args, **kwargs) -> T: ...
    def expm1(self, *args, **kwargs) -> T: ...
    def floor(self, *args, **kwargs) -> T: ...
    def floor_divide(self, *args, **kwargs) -> T: ...
    def greater(self, *args, **kwargs) -> T: ...
    def greater_equal(self, *args, **kwargs) -> T: ...
    def hypot(self, *args, **kwargs) -> T: ...
    def imag(self, *args, **kwargs) -> T: ...
    def isfinite(self, *args, **kwargs) -> T: ...
    def isinf(self, *args, **kwargs) -> T: ...
    def isnan(self, *args, **kwargs) -> T: ...
    def less(self, *args, **kwargs) -> T: ...
    def less_equal(self, *args, **kwargs) -> T: ...
    def log(self, *args, **kwargs) -> T: ...
    def log1p(self, *args, **kwargs) -> T: ...
    def log2(self, *args, **kwargs) -> T: ...
    def log10(self, *args, **kwargs) -> T: ...
    def logaddexp(self, *args, **kwargs) -> T: ...
    def logical_and(self, *args, **kwargs) -> T: ...
    def logical_not(self, *args, **kwargs) -> T: ...
    def logical_or(self, *args, **kwargs) -> T: ...
    def logical_xor(self, *args, **kwargs) -> T: ...
    def maximum(self, *args, **kwargs) -> T: ...
    def minimum(self, *args, **kwargs) -> T: ...
    def multiply(self, *args, **kwargs) -> T: ...
    def negative(self, *args, **kwargs) -> T: ...
    def nextafter(self, *args, **kwargs) -> T: ...
    def not_equal(self, *args, **kwargs) -> T: ...
    def positive(self, *args, **kwargs) -> T: ...
    def pow(self, *args, **kwargs) -> T: ...
    def real(self, *args, **kwargs) -> T: ...
    def reciprocal(self, *args, **kwargs) -> T: ...
    def remainder(self, *args, **kwargs) -> T: ...
    def round(self, *args, **kwargs) -> T: ...
    def sign(self, *args, **kwargs) -> T: ...
    def signbit(self, *args, **kwargs) -> T: ...
    def sin(self, *args, **kwargs) -> T: ...
    def sinh(self, *args, **kwargs) -> T: ...
    def square(self, *args, **kwargs) -> T: ...
    def sqrt(self, *args, **kwargs) -> T: ...
    def subtract(self, *args, **kwargs) -> T: ...
    def tan(self, *args, **kwargs) -> T: ...
    def tanh(self, *args, **kwargs) -> T: ...
    def trunc(self, *args, **kwargs) -> T: ...

    # Indexing Functions
    def take(self, *args, **kwargs) -> T: ...
    def take_along_axis(self, *args, **kwargs) -> T: ...

    # Inspection
    def __array_namespace_info__(self) -> ArrayNamespaceInfo: ...

    # Linear Algebra Functions
    def matmul(self, *args, **kwargs) -> T: ...
    def matrix_transpose(self, *args, **kwargs) -> T: ...
    def tensordot(self, *args, **kwargs) -> T: ...
    def vecdot(self, *args, **kwargs) -> T: ...

    # Manipulation Functions
    def broadcast_arrays(self, *args, **kwargs) -> list[T]: ...
    def broadcast_to(self, *args, **kwargs) -> T: ...
    def concat(self, *args, **kwargs) -> T: ...
    def expand_dims(self, *args, **kwargs) -> T: ...
    def flip(self, *args, **kwargs) -> T: ...
    def moveaxis(self, *args, **kwargs) -> T: ...
    def permute_dims(self, *args, **kwargs) -> T: ...
    def repeat(self, *args, **kwargs) -> T: ...
    def reshape(self, *args, **kwargs) -> T: ...
    def roll(self, *args, **kwargs) -> T: ...
    def squeeze(self, *args, **kwargs) -> T: ...
    def stack(self, arrays: tuple[T, ...] | list[T], /, *, axis: int = 0) -> T: ...
    def tile(self, *args, **kwargs) -> T: ...
    def unstack(self, *args, **kwargs) -> tuple[T, ...]: ...

    # Searching Functions
    def argmax(self, *args, **kwargs) -> T: ...
    def argmin(self, *args, **kwargs) -> T: ...
    def count_nonzero(self, *args, **kwargs) -> T: ...
    def nonzero(self, *args, **kwargs) -> tuple[T, ...]: ...
    def searchsorted(self, *args, **kwargs) -> T: ...
    def where(self, *args, **kwargs) -> T: ...

    # Set Functions
    def unique_all(self, *args, **kwargs) -> tuple[T, T, T, T]: ...
    def unique_counts(self, *args, **kwargs) -> tuple[T, T]: ...
    def unique_inverse(self, *args, **kwargs) -> tuple[T, T]: ...
    def unique_values(self, *args, **kwargs) -> T: ...

    # Sorting Functions
    def argsort(self, *args, **kwargs) -> T: ...
    def sort(self, *args, **kwargs) -> T: ...

    # Statistical Functions
    def cumulativ_prod(self, *args, **kwargs) -> T: ...
    def cumulativ_sum(self, *args, **kwargs) -> T: ...
    def max(self, *args, **kwargs) -> T: ...
    def mean(self, *args, **kwargs) -> T: ...
    def min(self, *args, **kwargs) -> T: ...
    def prod(self, *args, **kwargs) -> T: ...
    def std(self, *args, **kwargs) -> T: ...
    def sum(self, *args, **kwargs) -> T: ...
    def var(self, *args, **kwargs) -> T: ...

    # Utility Functions
    def all(self, *args, **kwargs) -> bool: ...
    def any(self, *args, **kwargs) -> bool: ...
    def diff(self, *args, **kwargs) -> T: ...



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends

class TestPolyval(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [120, 256, 280, 1024]

    def test_sin(self):
        vals = [100, 10, 1]
        offsets = [0.0, 0.5]
        for ts, size, val, offset in product(self.trainsum, self.sizes, vals, offsets):
            xp = ts.namespace
            dim = ts.dimension(size)
            domain = ts.domain(-1.0, 1.0)
            grid = ts.uniform_grid(dim, domain)

            x = xp.linspace(domain.lower, domain.upper, dim.size())
            exact = xp.sin(val*(x-offset))
            train = ts.sin(grid, val, offset)

            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2))
            self.assertLess(diff, 1e-7)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Protocol
from .backend import ArrayLike

class MatrixLeastSquares(Protocol):
    """Protocol for a matrix least squares solver."""
    
    def __call__(self, A: ArrayLike, B: ArrayLike, /) -> ArrayLike:
        """Solve the least squares problem for the given matrices A and B, returning the solution X that minimizes ||AX - B||."""
        ...





from typing import Sequence, Any
import unittest
from itertools import product
from copy import deepcopy

from trainsum import TrainSum
from trainsum.typing import UniformGrid, TrainShape
from utils import backends, rand_data

class TestTensorTrain(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120,), (280,), (1024,), (120, 1024), (324, 120)]

    def get_grid(self, ts, sizes: Sequence[int], lower: float, upper: float):
        dims = [ts.dimension(size) for size in sizes]
        domains = [ts.domain(lower, upper) for _ in sizes]
        return ts.uniform_grid(dims, domains)

    def get_idxs(self, ts, grid: UniformGrid):
        xp = ts.namespace
        idxs = xp.zeros([len(grid.dims), *[dim.size() for dim in grid.dims]],
                        dtype=ts.index_type)
        for i, dim in enumerate(grid.dims):
            cut = (*(xp.newaxis,) * i, slice(None), *(xp.newaxis,) * (len(grid.dims) - i - 1))
            idxs[i] += xp.arange(dim.size(), dtype=ts.index_type)[cut]
        return idxs
    
    def rand_cores(self, ts: TrainSum, shape: TrainShape):
        xp = ts.namespace
        cores = []
        for i in range(len(shape)):
            left = 1 if i == 0 else 10
            right = 1 if i == len(shape) - 1 else 10
            cores.append(xp.asarray(rand_data(xp, left, *shape.middle(i), right)))
        return cores

    def left_contract(self, ts: TrainSum, train: Any, idx: int):
        xp = ts.namespace
        tmp = xp.ones((1, 1))
        for i in range(idx):
            idxs = [i+1 for i in range(len(train.shape.middle(i)))]
            mid = xp.tensordot(train.cores[i], train.cores[i], axes=(idxs, idxs))
            tmp = xp.tensordot(tmp, mid, axes=([0, 1], [0, 2]))
        return tmp

    def right_contract(self, ts: TrainSum, train: Any, idx: int):
        xp = ts.namespace
        tmp = xp.ones((1, 1))
        for i in range(len(train.shape)-1, idx+1, -1):
            idxs = [i+1 for i in range(len(train.shape.middle(i)))]
            mid = xp.tensordot(train.cores[i], train.cores[i], axes=(idxs, idxs))
            tmp = xp.tensordot(mid, tmp, axes=([1, 3], [0, 1]))
        return tmp


    def test_extend(self) -> None:
        for ts, sizes1, sizes2 in product(self.trainsum, self.sizes, self.sizes):
            xp = ts.namespace

            shape1 = ts.trainshape(*sizes1, mode="block")
            cores1 = self.rand_cores(ts, shape1)
            train1 = ts.tensortrain(shape1, cores1)

            shape2 = ts.trainshape(*sizes2, mode="interleaved")
            cores2 = self.rand_cores(ts, shape2)
            train2 = ts.tensortrain(shape2, cores2)

            res = deepcopy(train1)
            res.extend(train2)
            core_iter = iter(res.cores)

            self.assertEqual(list(train1.shape.dims) + list(train2.shape.dims), res.shape.dims)
            for core in train1.cores:
                self.assertTrue(xp.all(xp.equal(core, next(core_iter))))
            for core in train2.cores:
                self.assertTrue(xp.all(xp.equal(core, next(core_iter))))

    def test_conj(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace
            ctype = xp.__array_namespace_info__().dtypes()["complex128"]

            shape = ts.trainshape(*sizes, mode="block")
            cores = self.rand_cores(ts, shape)
            train = ts.tensortrain(shape, cores)
            train.dtype = ctype

            res = train.conj()
            core_iter = iter(res.cores)

            self.assertEqual(train.shape.dims, res.shape.dims)
            for core in train.cores:
                self.assertTrue(xp.all(xp.equal(xp.conj(core), next(core_iter))))

    def test_normalize(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace

            shape = ts.trainshape(*sizes, mode="block")
            cores = self.rand_cores(ts, shape)
            train = ts.tensortrain(shape, cores)

            for i in range(len(shape)):
                train.normalize(i)

                left = self.left_contract(ts, train, i)
                eye_left = xp.eye(left.shape[0])
                diff = xp.sum((left - eye_left)**2)
                self.assertLess(diff, 1e-7)

                right = self.right_contract(ts, train, i)
                eye_right = xp.eye(right.shape[0])
                diff = xp.sum((right - eye_right)**2)
                self.assertLess(diff, 1e-7)

    def test_truncate(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            shape = ts.trainshape(*sizes, mode="block")
            cores = self.rand_cores(ts, shape)
            train = ts.tensortrain(shape, cores)

            with ts.decomposition(max_rank=5):
                train.truncate()
            self.assertLessEqual(max(train.shape.ranks), 5)

            with ts.variational(max_rank=5):
                train.truncate()
            self.assertLessEqual(max(train.shape.ranks), 5)

    def test_transform(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -10.0, 10.0)
            idxs = self.get_idxs(ts, grid)
            coords = grid.to_coords(idxs)

            data = xp.sum(coords**2, axis=0)
            shape = ts.trainshape(*sizes, mode="block")
            with ts.variational(max_rank=15, cutoff=1e-10, nsweeps=1, ncores=2):
                train = ts.tensortrain(shape, data)

            func = lambda x: x**2
            with ts.cross(max_rank=32, eps=1e-10):
                res = train.transform(func)

            exact = func(train.to_tensor())
            approx = res.to_tensor()
            diff = xp.sum((exact - approx)**2) / xp.sum(exact**2)
            self.assertLess(diff, 1e-5)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

class Domain:                                                                                                                       
    """Domnain of a variable, defined by a lower and upper bound."""

    _lower: float
    _upper: float
    _diff: float

    @property
    def lower(self) -> float:
        return self._lower

    @property
    def upper(self) -> float:
        return self._upper

    @property
    def diff(self) -> float:
        return self._diff

    def __init__(self, lower: float, upper: float):
        if lower >= upper:
            raise ValueError("Upper bound must be greater than lower bound")
        self._lower = lower
        self._upper = upper
        self._diff = upper - lower

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Domain)\
               and self.lower == other.lower\
               and self.upper == other.upper
    
    def __str__(self) -> str:
        return f"Domain({self.lower}, {self.upper})"



from typing import Sequence, Any
import unittest
from itertools import product

from trainsum import TrainSum
from trainsum.typing import UniformGrid, TrainShape
from utils import backends, rand_data

class TestConstruct(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]

    def get_grid(self, ts, sizes: Sequence[int], lower: float, upper: float):
        dims = [ts.dimension(size) for size in sizes]
        domains = [ts.domain(lower, upper) for _ in sizes]
        return ts.uniform_grid(dims, domains)

    def get_idxs(self, ts, grid: UniformGrid):
        xp = ts.namespace
        idxs = xp.zeros([len(grid.dims), *[dim.size() for dim in grid.dims]],
                        dtype=ts.index_type)
        for i, dim in enumerate(grid.dims):
            cut = (*(xp.newaxis,) * i, slice(None), *(xp.newaxis,) * (len(grid.dims) - i - 1))
            idxs[i] += xp.arange(dim.size(), dtype=ts.index_type)[cut]
        return idxs

    def check_data_construct(self, ts: TrainSum, shape: TrainShape, data: Any) -> None:
        xp = ts.namespace
        train = ts.tensortrain(shape, data)
        approx = train.to_tensor()
        diff = abs(xp.sum((data - approx)**2))
        self.assertLess(diff, 1e-5)

    def check_cross_construct(self, ts: TrainSum, shape: TrainShape, grid: UniformGrid, func: Any) -> None:
        xp = ts.namespace
        train = ts.tensortrain(shape, func)

        idxs = self.get_idxs(ts, grid)
        exact = func(idxs)

        approx = train.to_tensor()
        diff = abs(xp.sum((exact - approx)**2))
        self.assertLess(diff, 1e-5)

    def test_data(self):
        sizes = [(1024,), (120, 1024), (324, 120)]
        for ts, sizes in product(self.trainsum, sizes):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -10.0, 10.0)
            idxs = self.get_idxs(ts, grid)
            coords = grid.to_coords(idxs)

            exact = xp.exp(-0.5 * xp.sum(coords**2, axis=0))
            shape = ts.trainshape(*grid.dims, mode="block")
            with ts.exact():
                self.check_data_construct(ts, shape, exact)
            with ts.decomposition(max_rank=15, cutoff=1e-10, ncores=2):
                self.check_data_construct(ts, shape, exact)
            with ts.variational(max_rank=15, cutoff=1e-10, nsweeps=1, ncores=2):
                self.check_data_construct(ts, shape, exact)

    def test_func(self):
        sizes = [(1024,), (120, 1024), (324, 120)]
        for ts, sizes in product(self.trainsum, sizes):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -10.0, 10.0)

            func = lambda idxs: xp.exp(-0.5 * xp.sqrt(xp.sum(grid.to_coords(idxs)**2, axis=0)))
            shape = ts.trainshape(*grid.dims, mode="block")
            with ts.cross(max_rank=32, eps=1e-10):
                self.check_cross_construct(ts, shape, grid, func)

    def test_explicit(self):
        sizes = [(1024,), (120, 1024), (324, 120)]
        for ts, sizes in product(self.trainsum, sizes):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -10.0, 10.0)

            shape = ts.trainshape(*grid.dims, mode="interleaved")
            cores = []
            for i in range(len(shape)):
                left = 1 if i == 0 else 16
                right = 1 if i == len(shape) - 1 else 16
                cores.append(rand_data(xp, left, *shape.middle(i), right))
            train = ts.tensortrain(shape, cores)
            for ref_core, core in zip(cores, train.cores):
                self.assertTrue(xp.all(xp.equal(ref_core, core)))

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike, namespace_of_arrays, shape
from .trainshape import TrainShape
from .trainbase import TrainBase

def to_train[T: ArrayLike](tshape: TrainShape, data: T) -> TrainBase[T]:
    xp = namespace_of_arrays(data)
    if len(tshape.dims) != len(shape(data)):
        raise ValueError("Data dimensionality does not match the provided TrainShape.")
    if any(dim.size() != s for dim, s in zip(tshape.dims, shape(data))):
        raise ValueError("Data shape does not match the provided TrainShape.")
    ref_digits = []
    for dgts in tshape.digits:
        ref_digits.extend(dgts)
    tshape = TrainShape(tshape.dims, [ref_digits])

    digits = [d for dim in tshape.dims for d in dim]
    ref_digits = sum((list(digits) for digits in tshape.digits), start=[])
    perm = [digits.index(d) for d in ref_digits]

    data = xp.reshape(data, [d.base for dim in tshape.dims for d in dim])
    data = xp.permute_dims(data, perm)
    return TrainBase(tshape, [data[xp.newaxis,...,xp.newaxis]])



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from math import prod

from .localcontraction import LocalContraction
from .backend import ArrayLike, shape_size, size, namespace_of_arrays
from .dimension import Dimension
from .digit import Digit, Digits
from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction, get_symbol_generator
from .contractorinput import ContractorInput
from .utils import check_operand_shapes, get_shapes, shape_map
from .coreslicing import CoreSlicing
from .contractor import ArrayContractor, OptimizeKind

class Evaluation:

    optimizer: OptimizeKind
    chunk_size: int
    _contr: EinsumContraction
    _inp: None | ContractorInput
    _tmp_str: str
    _exprs: Sequence[ArrayContractor]
    _core_slicing: Sequence[CoreSlicing]
    _res_digits: Sequence[Digits]

    def __init__(
            self,
            contr: EinsumContraction,
            optimizer: OptimizeKind = "greedy",
            chunk_size: int = 1024) -> None:
        self._contr = contr
        self.optimizer = optimizer
        self.chunk_size = chunk_size
        self._tmp_str = next(get_symbol_generator(contr))

        self._inp = None
        self._exprs = []
        self._core_slicing = []
        self._res_digits = []

    def __call__[T: ArrayLike](
            self,
            idxs: T,
            *ops: TrainBase[T],
            expr: bool = False) -> T:
        #idxs_ = {dim: idxs[char] for char, dim in zip(self._contr.equation.result, self._contr.equation.result_dims)}
        idxs_ = {dim: idxs[i] for i, dim in enumerate(self._contr.equation.result_dims)}
        self._check_idxs(idxs_)

        shapes = get_shapes(*ops)
        if expr or self._inp is None:
            self.calc_expressions(*shapes)
        else:
            self._inp.check_operands(*ops)

        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        xp, device, dtype = self._inp.infos(*ops)
        dmap = _digit_map(idxs_)

        shape = next(iter(idxs_.values())).shape
        length = shape_size(shape)
        res = xp.empty(length, device=device, dtype=dtype)
        for i in range(length // self.chunk_size + 1):
            start = i * self.chunk_size
            stop = min((i+1) * self.chunk_size, length)
            if start >= stop:
                break
            chunk_idxs = {digit: vals[start:stop] for digit, vals in dmap.items()}
            res[start:stop] = self._contract(chunk_idxs, *ops)

        return xp.reshape(res, shape)

    def calc_expressions(self, *operands: TrainShape | TrainBase) -> None:
        shapes = get_shapes(*operands)
        check_operand_shapes(self._contr.operand_shapes, shapes)

        self._const_ops = [op for op in operands if not isinstance(op, TrainShape)]
        self._inp = ContractorInput(*operands)

        self._exprs = []
        self._core_slicing = []
        self._res_digits = []
        for lcontr in self._contr:
            self._exprs.append(self._expression(lcontr, *shapes))
            self._core_slicing.append(CoreSlicing(lcontr, *operands))
            self._res_digits.append(_result_digits(lcontr, *self._contr.operand_shapes))

    # ------------------------------------------------------------------------
    # Contraction generators

    def _contract[T: ArrayLike](
            self,
            dmap: dict[Digit, T],
            *ops: TrainBase[T]) -> T:
        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        xp, device, dtype = self._inp.infos(*ops)
        length = size(next(iter(dmap.values())))

        res = xp.ones((length, *[1]*len(self._contr[0].result.left)),
                      device=device, dtype=dtype)
        tmp = xp.empty(length, device=device, dtype=dtype)
        for expr, slc, digits in zip(self._exprs, self._core_slicing, self._res_digits):
            new_res = xp.empty((length, *slc.get_right_shape(*ops)),
                               device=device, dtype=dtype)
            tmp = _get_ids(dmap, digits, tmp)
            for i in range(prod(d.base for d in digits)):
                mask = xp.where(tmp == i, True, False)
                if size(mask) == 0:
                    continue
                mats = slc.get_matrices(i, *ops)
                new_res[mask,...] = expr(res[mask], *mats)
            res = new_res

        return res[:,0]

    # ------------------------------------------------------------------------
    # Expression builders

    def _expression(self, lcontr: LocalContraction, *shapes: TrainShape) -> ArrayContractor:
        tmp = f"{self._tmp_str}{lcontr.result.left}"
        res = f"{self._tmp_str}{lcontr.result.right}"
        ops = []
        tns = []
        for op, tn in zip(lcontr.operands, lcontr.get_shapes(*shapes)):
            new_mid = ""
            new_shape = [tn[0]]
            for i, char in enumerate(op.middle):
                if char in lcontr.result.middle:
                    continue
                new_mid += char
                new_shape.append(tn[i+1])
            new_shape.append(tn[-1])
            ops.append(f"{op.left}{new_mid}{op.right}")
            tns.append(new_shape)

        eq = f"{tmp}," + ",".join(op for op in ops) + f"->{res}"

        smap = shape_map(shapes, lcontr)
        smap[self._tmp_str] = self.chunk_size
        tmp_shape = [smap[char] for char in tmp]

        return ArrayContractor(eq, tmp_shape, *tns,
                               optimizer=self.optimizer)

    def _check_idxs[T: ArrayLike](self, idxs: dict[Dimension, T]) -> None:
        if not all(dim in self._contr.equation.result_dims for dim in idxs.keys()):
            raise ValueError("All dimensions in idxs must be in the result dimensions of the equation.")
        if len(set(idxs[dim].shape for dim in idxs.keys())) != 1:
            raise ValueError("All index arrays must have the same .")

def _digit_map[T: ArrayLike](idxs: dict[Dimension, T]) -> dict[Digit, T]:
    xp = namespace_of_arrays(next(iter(idxs.values())))
    dmap = {}
    for dim, didxs in idxs.items():
        vals = dim.to_digits(didxs)
        for i, digit in enumerate(dim):
            length = size(vals[i])
            dmap[digit] = xp.reshape(vals[i,...], (length,))
    return dmap

def _get_ids[T: ArrayLike](dmap: dict[Digit, T], digits: Digits, x: T) -> T:
    ncombs = prod(digit.base for digit in digits)
    x[...] = 0
    for digit in digits:
        ncombs //= digit.base
        x += dmap[digit]*ncombs
    return x

def _result_digits(lcontr: LocalContraction, *ops: TrainShape) -> Digits:
    dmap = {}
    for op_str, tidx, cidx in zip(lcontr.operands, lcontr.train_idxs, lcontr.core_idxs):
        digits = ops[tidx].digits[cidx]
        for char, digit in zip(op_str.middle, digits):
            if digit not in dmap:
                dmap[char] = digit
    return [dmap[char] for char in lcontr.result.middle]



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from types import NoneType
from typing import Callable
from .backend import ArrayLike, ArrayNamespace
from .options import get_options, OptionType, ExactOptions, DecompositionOptions, VariationalOptions
from .trainshape import TrainShape
from .trainbase import TrainBase
from .utils import namespace_of_trains
from .einsumexpression import EinsumExpression as _EinsumExpression

class EinsumExpression[T: ArrayLike]:

    _expr: Callable

    def __init__(
        self,
        xp: ArrayNamespace[T],
        eq: str,
        *ops: TrainShape,
        result_shape: NoneType | TrainShape = None,
        ) -> None:
        opts = get_options(xp, OptionType.EINSUM)
        if isinstance(opts, ExactOptions):
            self._expr = _EinsumExpression(
                    eq, *ops,
                    method="exact",
                    optimizer=opts.optimizer)
        elif isinstance(opts, DecompositionOptions):
            self._expr = _EinsumExpression(
                    eq, *ops,
                    method="decomposition",
                    decomposition=opts.decomposition,
                    strategy=opts.strategy,
                    optimizer=opts.optimizer,
                    result_shape=result_shape)
        elif isinstance(opts, VariationalOptions):
            self._expr = _EinsumExpression(
                    eq, *ops,
                    method="variational",
                    decomposition=opts.decomposition,
                    strategy=opts.strategy,
                    optimizer=opts.optimizer,
                    result_shape=result_shape)
        else:
            raise RuntimeError("Invalid options type.")

    def __call__(
            self,
            *ops: TrainBase[T],
            ) -> float | TrainBase[T]:
        return self._expr(*ops)

def einsum[T: ArrayLike](
        eq: str,
        *ops: TrainBase[T],
        result_shape: NoneType | TrainShape = None,
        ) -> float | TrainBase[T]:
    xp = namespace_of_trains(*ops)
    expr = EinsumExpression(xp, eq, *[op.shape for op in ops], result_shape=result_shape)
    return expr(*ops)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from copy import deepcopy

from .backend import ArrayLike
from .direction import Direction
from .sweepingstrategy import SweepingStrategy
from .trainbase import TrainBase
from .matrixdecomposition import MatrixDecomposition, MatrixDecompositionResult
from .svdecomposition import SVDecomposition
from .utils import trains_match
from .tensordecomposition import TensorDecomposition
from .additioncore import AdditionCore

class DecompositionAddition[T: ArrayLike, S: MatrixDecompositionResult]:

    @property
    def decomposition(self) -> MatrixDecomposition[T, S]:
        return self._mat_decomp
    @decomposition.setter
    def decomposition(self, value: MatrixDecomposition[T, S]) -> None:
        self._mat_decomp = deepcopy(value)
        self._tn_decomp = TensorDecomposition(value)
    
    strategy: SweepingStrategy
    _mat_decomp: MatrixDecomposition[T, S]
    _tn_decomp: TensorDecomposition[T, S]

    def __init__(
            self,
            strategy: SweepingStrategy,
            decomposition: MatrixDecomposition[T, S] = SVDecomposition[T]()) -> None:
        self.strategy = deepcopy(strategy)
        self.decomposition = deepcopy(decomposition)

    def __call__(
            self,
            train1: TrainBase[T],
            train2: TrainBase[T],
            *trains: TrainBase[T],
            direction: Direction = Direction.TO_RIGHT,
            ) -> TrainBase[T]:
        trains = (train1, train2, *trains)
        trains_match(*trains)
        if direction == Direction.TO_RIGHT:
            [train.normalize(0) for train in trains]
            return self._to_right(*trains)
        else:
            [train.normalize(-1) for train in trains]
            return self._to_left(*trains)

    def _to_right(self, *trains: TrainBase[T]) -> TrainBase[T]:
        shape = trains[0].shape
        acore = AdditionCore[T](shape, Direction.TO_RIGHT)
        lranges = self.strategy.right_sweep(shape)

        data = []
        begin, end = lranges[0].begin, 0
        for lrange in lranges:
            for i in range(begin, lrange.begin):
                tmp = acore.cut(self._tn_decomp)
                data.append(tmp.data)
            for i in range(max(end, lrange.begin), lrange.end):
                acore.add(i, *trains)
            begin, end = lrange.begin, lrange.end

        for i in range(begin, len(shape)):
            tmp = acore.cut(self._tn_decomp)
            data.append(tmp.data)

        return TrainBase(shape, data, copy_data=False)

    def _to_left(self, *trains: TrainBase[T]) -> TrainBase[T]:
        shape = trains[0].shape
        acore = AdditionCore[T](shape, Direction.TO_LEFT)
        lranges = self.strategy.left_sweep(shape)

        data = []
        begin, end = lranges[0].end, lranges[0].end
        for lrange in lranges:
            for i in range(end-1, lrange.end-1, -1):
                tmp = acore.cut(self._tn_decomp)
                data.append(tmp.data)
            for i in range(min(begin, lrange.end)-1, lrange.begin-1, -1):
                acore.add(i, *trains)
            begin, end = lrange.begin, lrange.end

        for i in range(end-1, -1, -1):
            tmp = acore.cut(self._tn_decomp)
            data.append(tmp.data)
        data.reverse()

        return TrainBase(shape, data, copy_data=False)



import unittest
from itertools import product
from scipy.linalg import dft

from trainsum import TrainSum
from utils import backends

class TestQft(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [120, 256, 280, 1024]

    def test_qft(self):
        for ts, size in product(self.trainsum, self.sizes):
            xp = ts.namespace
            dim = ts.dimension(size)
            decomp = ts.svdecomposition(max_rank=16, cutoff=0.0)

            exact = xp.asarray(dft(dim.size(), "sqrtn"))
            train = ts.qft(dim, decomp)

            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2) / xp.sum(exact**2))
            self.assertLess(diff, 1e-4)

    def test_iqft(self):
        for ts, size in product(self.trainsum, self.sizes):
            xp = ts.namespace
            dim = ts.dimension(size)
            decomp = ts.svdecomposition(max_rank=16, cutoff=0.0)

            exact = xp.conj(xp.asarray(dft(dim.size(), "sqrtn"))).T
            train = ts.iqft(dim, decomp)

            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2) / xp.sum(exact**2))
            self.assertLess(diff, 1e-4)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from math import exp as mexp

from .backend import ArrayNamespace, ArrayLike
from .uniformgrid import UniformGrid
from .trainbase import TrainBase
from .trainshape import TrainShape

def exp[T: ArrayLike](xp: ArrayNamespace[T], grid: UniformGrid, factor: float, offset: float) -> TrainBase[T]:
    if len(grid.dims) != 1:
        raise ValueError("Exponential function only supports 1D uniform grid")
    dim = grid.dims[0]
    shape = TrainShape(dim, [(d,) for d in dim])
    domain = grid.domains[0]
    spacing = grid.spacings[0]
    
    data = []
    for digit in dim:
        xi = spacing*digit.factor
        tmp = xp.zeros((1, digit.base, 1))
        for j in range(digit.base):
            tmp[:,j,:] = mexp(factor * j*xi)
        data.append(tmp)
    data[0] *= mexp(factor * (domain.lower - offset))
    return TrainBase(shape, data, copy_data=False)



import unittest
from itertools import product
import os
import h5py

from trainsum import TrainSum
from trainsum.typing import TensorTrain, UniformGrid
from utils import backends, rand_data

path = os.path.dirname(__file__)
class TestIO(unittest.TestCase):

    def setUp(self) -> None:
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120,), (256,18), (280,), (1024,20)]
        self.modes = ("block", "interleaved")

        os.mkdir(f"{path}/data")
        self.file = h5py.File("data/test_io.h5", "w")

    def tearDown(self) -> None:
        self.file.close()
        os.remove(f"{path}/data/test_io.h5")
        os.rmdir(f"{path}/data")

    def test_tensortrain(self) -> None:
        for ts, sizes, mode in product(self.trainsum, self.sizes, self.modes):
            xp = ts.namespace
            name = f"tensortrain_{sizes}"
            group = self.file.create_group(name)

            shape = ts.trainshape(*sizes, mode=mode)
            cores = []
            for i in range(len(shape)):
                left = 1 if i == 0 else 10
                right = 1 if i == len(shape)-1 else 10
                cores.append(rand_data(xp, left, *shape.middle(i), right))
            ref_train = ts.tensortrain(shape, cores)

            ts.write(group, ref_train)
            train = ts.read(group, TensorTrain)

            self.assertEqual(train.shape, ref_train.shape)
            for core, ref_core in zip(train.cores, ref_train.cores):
                self.assertTrue(xp.all(xp.equal(core, ref_core)))

            del self.file[name]

    def test_uniformgrid(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            name = f"uniformgrid_{sizes}"
            group = self.file.create_group(name)

            dims = [ts.dimension(size) for size in sizes]
            domains = [ts.domain(0, 1) for _ in sizes]
            ref_grid = ts.uniform_grid(dims, domains)

            ts.write(group, ref_grid)
            grid = ts.read(group, UniformGrid)

            self.assertEqual(grid, ref_grid)
            for domain, ref_domain in zip(grid.domains, ref_grid.domains):
                self.assertEqual(domain, ref_domain)

            del self.file[name]

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from copy import deepcopy
from typing import Sequence
from dataclasses import dataclass, field

from .backend import ArrayLike
from .operandstring import OperandString
from .trainshape import TrainShape
from .trainbase import TrainBase


@dataclass(frozen=True)
class LocalContraction:

    operands: Sequence[OperandString] = field(default_factory=list)
    result: OperandString = field(default_factory=OperandString)

    train_idxs: Sequence[int] = field(default_factory=list)
    core_idxs: Sequence[int] = field(default_factory=list)

    def add_operand(self, train_idx: int, core_idx: int, op_str: OperandString) -> None:
        object.__setattr__(self, "train_idxs", [*self.train_idxs, train_idx])
        object.__setattr__(self, "core_idxs", [*self.core_idxs, core_idx])
        object.__setattr__(self, "operands", [*self.operands, op_str])

    def set_result(self, res_str: OperandString) -> None:
        object.__setattr__(self, "result", res_str)

    def get_data[T: ArrayLike](self,
                 *trains: TrainBase[T],
                 idx_map: dict[int, int] = {}) -> Sequence[T]:
        """Get the data arrays for the operands involved in this contraction."""
        out = []
        for op_idx, core_idx in zip(self.train_idxs, self.core_idxs):
            if op_idx in idx_map:
                idx = idx_map[op_idx]
                out.append(trains[idx].data[core_idx])
        return out

    def get_shapes(self, *ops: TrainShape | TrainBase) -> Sequence[Sequence[int]]:
        """Get the shapes of the data tensors for the operands involved in this contraction."""
        out = []
        for op_idx, core_idx in zip(self.train_idxs, self.core_idxs):
            op = ops[op_idx]
            if not isinstance(op, TrainShape):
                out.append(op.data[core_idx].shape)
            else:
                out.append((op.left_rank(core_idx), *op.middle(core_idx), op.right_rank(core_idx)))
        return out

    def get_constants[T: ArrayLike](self,
                      *ops: TrainShape | TrainBase[T]
                      ) -> tuple[Sequence[int], Sequence[Sequence[int] | T]]:
        idxs = []
        out = []
        for i, (op_idx, core_idx) in enumerate(zip(self.train_idxs, self.core_idxs)):
            op = ops[op_idx]
            if isinstance(op, TrainShape):
                out.append([op.left_rank(core_idx),
                            *op.middle(core_idx),
                            op.right_rank(core_idx)])
            else:
                out.append(op.data[core_idx])
                idxs.append(i)
        return idxs, out

    def __str__(self) -> str:
        ops_str = ",".join(str(op) for op in self.operands)
        res_str = str(self.result)
        return f"{ops_str} -> {res_str}"

def fuse_local_contractions(ops: Sequence[LocalContraction]) -> LocalContraction:
    if len(ops) == 0:
        raise ValueError("Cannot fuse zero LocalContractions.")
    elif len(ops) == 1:
        return ops[0]
    contr = deepcopy(ops[0])
    res = ops[0].result
    for i in range(1, len(ops)):
        prev_right = ops[i-1].result.right
        cur_left = ops[i].result.left
        if set(prev_right) != set(cur_left):
            raise ValueError("LocalContractions have incompatible contraction indices.")
        res = OperandString(left=res.left,
                            middle=res.middle+ops[i].result.middle,
                            right=ops[i].result.right)
        for op_str, train_idx, core_idx in zip(ops[i].operands, ops[i].train_idxs, ops[i].core_idxs):
            contr.add_operand(train_idx, core_idx, op_str)
    contr.set_result(res)
    return contr



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any, Sequence, Callable
from dataclasses import dataclass
from copy import deepcopy
import time


from .matrixeigenvaluedecomposition import MatrixEigenvalueDecomposition
from .eighsolver import EigHSolver
from .backend import DType, Device, ArrayLike, namespace_of_arrays, size, shape

class LanczosData[T: ArrayLike]:
    device: Device
    dtype: DType
    subspace: int
    basis: T
    offset: int
    eigvec: T
    eigval: float

    def __init__(self, subspace: int, guess: T, states: Sequence[T]) -> None:
        self._check_input(guess, states)
        self.device = guess.device
        self.dtype = guess.dtype
        self.subspace = min(subspace, size(guess)-len(states))

        xp = namespace_of_arrays(guess)
        self.basis = xp.zeros((self.subspace+len(states), *guess.shape), 
                              device=guess.device, dtype=guess.dtype)
        self.offset = len(states)
        for i, state in enumerate(states):
            self.basis[i,:] = state
        self.eigvec = deepcopy(guess)

    def _check_input(self, guess: T, states: Sequence[T]) -> None:
        if not all(guess.device == state.device for state in states):
            raise ValueError("All inputs must be on the same device")
        if not all(guess.dtype == state.dtype for state in states):
            raise ValueError("All inputs must have the same dtype")

@dataclass(kw_only=True)
class LanczosResult[T: ArrayLike]:
    #: Eigenvector
    array: T
    #: Eigenvalue
    value: float
    #: Time taken to compute the eigenvector and eigenvalue.
    time: float
    #: Convergence history of the eigenvalues at each step.
    eps: list[float]

@dataclass
class Lanczos:
    """
    Lanczos eigenvalue solver.
    """

    #: Number of times the subspace is created and solved
    nsteps: int = 1

    #: Size of the Lanczos basis
    subspace: int = 3

    #: Minimum difference between eigenvalues of consecutive steps, after which the algorithm is stopped
    eps: float = 1e-8

    #: Eigenvalue solver for diagonalizing the matrix in the Lanczos basis
    solver: MatrixEigenvalueDecomposition = EigHSolver()

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ("eps", "nsteps", "subspace") and value <= 0:
            raise ValueError(f"{name} must be a positive number above zero.")
        super().__setattr__(name, value)

    def __call__[T: ArrayLike](
            self,
            ham: Callable[[T], T],
            guess: T,
            states: Sequence[T] = [], /) -> LanczosResult[T]:
        """
        Solve the eigenvalue problem for some linear map with an initial guess and a set of orthogonal states.
        """
        xp = namespace_of_arrays(guess)
        self._check_input(guess, states)

        if min(self.subspace, size(guess)-len(states)) < len(states):
            return LanczosResult(array=guess,
                                 value=float(xp.sum(guess*ham(guess))),
                                 time=0.0,
                                 eps=[0.0])


        stamp = time.time()

        data = LanczosData(self.subspace, guess, states)
        data.eigvec = deepcopy(guess)
        for i in range(data.offset):
            self._gram_schmidt(data.basis[i,:], data.basis[:i,:])
            data.basis[i,:] /= xp.sqrt(xp.sum(data.basis[i,:]*data.basis[i,:]))

        eps = [float(xp.sum(data.eigvec*ham(data.eigvec)))]
        for _ in range(self.nsteps):
            self._step(ham, data)
            eps.append(data.eigval)
            if abs(eps[-1]-eps[-2]) < self.eps:
                break

        return LanczosResult(array=data.eigvec,
                             value=float(xp.sum(data.eigvec*ham(data.eigvec))),
                             time=time.time() - stamp,
                             eps=eps)
        
    def _step[T: ArrayLike](
            self,
            mat: Callable[[T], T],
            data: LanczosData[T]) -> None:

        xp = namespace_of_arrays(data.basis)
        vecs, guess, subspace, off = data.basis, data.eigvec, data.subspace, data.offset

        a = xp.zeros(subspace, dtype=data.dtype)
        b = xp.zeros(subspace, dtype=data.dtype)
        sub_mat = xp.zeros((data.subspace, data.subspace))

        vecs[off] = guess
        for j in range(off):
            val = xp.sum(vecs[off,:] * vecs[j,:])
            vecs[off,:] = vecs[off,:] - val * vecs[j,:]
        vecs[off] /= xp.sqrt(xp.sum(vecs[off]*vecs[off]))

        for i in range(off+1, off+subspace):
            vecs[i,:] = mat(vecs[i-1,:])
            if i > off+1:
                vecs[i,:] = vecs[i,:] - b[i-off-2] * vecs[i-2,:]
            a_val = xp.sum(vecs[i-1,:] * vecs[i,:])
            a[i-off-1] = a_val
            vecs[i,:] = vecs[i,:] - a_val * vecs[i-1,:]
            self._gram_schmidt(vecs[i,:], vecs[:i,:])

            b_val = xp.sqrt(xp.sum(vecs[i,:] * vecs[i,:]))
            vecs[i,:] = vecs[i,:] / b_val
            b[i-off-1] = b_val
        a[-1] = xp.sum(vecs[-1,:] * mat(vecs[-1,:]))

        sub_mat[0,0] = a[0]
        sub_mat[0,1] = b[0]
        for i in range(1, data.subspace-1):
            sub_mat[i,i-1] = b[i-1]
            sub_mat[i,i] = a[i]
            sub_mat[i,i+1] = b[i]
        sub_mat[-1,-1] = a[-1]
        sub_mat[-1,-2] = b[-2]

        eigvals, eigvecs = self.solver(sub_mat)
        data.eigval = float(eigvals[0])
        data.eigvec = xp.tensordot(eigvecs[:,0], vecs[off:,...], axes=([0],[0]))

    def _gram_schmidt[T: ArrayLike](self, vec: T, basis: T) -> None:
        xp = namespace_of_arrays(vec)
        for i in range(shape(basis)[0]):
            val = xp.sum(vec * basis[i,:])
            vec -= val * basis[i,:]

    def _check_input[T: ArrayLike](self, guess: T, states: Sequence[T]) -> None:
        if any(orth.shape != guess.shape for orth in states):
            raise ValueError("Orthogonal states must have the same dimensions as the initial state")



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

import cupy as cp
from .trainsum import TrainSum

trainsum = TrainSum[cp.ndarray](cp)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence, overload
from dataclasses import dataclass
from copy import deepcopy
from math import prod
from types import NoneType

from .backend import ArrayLike, namespace_of_arrays, device, DType
from .digit import Digit, digits_similar
from .sequenceof import SequenceOf
from .backend import get_index_dtype

@dataclass(frozen=True, init=False)
class Dimension(SequenceOf[Digit]):
    """
    A quantized dimension, represented as a sequence of digits. If a integer is provided during initialization,
    it will be factorized into its prime factors to create the digits. Otherwise, the provided sequence of integers
    will be used as the bases of the digits.
    """

    #-------------------------------------------------------------------------
    #private members

    #: The identifier of the dimension, shared by all digits in the dimension.
    idf: int

    #-------------------------------------------------------------------------
    #constructor

    @overload
    def __init__(self,
                 size: int, /,
                 idf: int | NoneType = None) -> None: ...
    @overload
    def __init__(self,
                 bases: Sequence[int], /,
                 idf: int | NoneType = None) -> None: ...
    def __init__(self,
                 size: int | Sequence[int], /,
                 idf: int | NoneType = None) -> None:
        if isinstance(size, int):
            bases = prime_factorization(size)
        else:
            bases = size

        if isinstance(idf, NoneType):
            idf = unique_identifier()
        object.__setattr__(self, "idf", idf)

        self._check_bases(bases)
        factors = [prod(bases[i+1:]) for i in range(len(bases))]
        digits = []
        for i, base, factor in zip(range(len(bases)), bases, factors):
            digits.append(Digit(idf, i, base, factor))
        super().__init__(digits)

    def _check_bases(self, bases: Sequence[int]) -> None:
        if len(bases) == 0:
            raise ValueError("Dimension must have at least one digit")
        if any(b < 2 for b in bases):
            raise ValueError("All bases must be at least 2")

    #-------------------------------------------------------------------------
    #methods

    def size(self) -> int:
        """Calculate the size of the dimension, i.e., the product of the bases of all digits."""
        return prod(d.base for d in self)

    def to_idxs[T: ArrayLike](self, digits: T) -> T:
        """
        Convert digit indices with shape (len(Dimension), ...) to dimension indices with shape (...).
        """
        xp = namespace_of_arrays(digits)
        int_type = get_index_dtype(xp)
        self._check_dtype(int_type, digits)
        if digits.shape[0] != len(self):
            raise ValueError(f"Expect a tensor of shape ({len(self)}, ...)")
        trans = xp.asarray([d.factor for d in self],
                            dtype=int_type,
                            device=device(digits))
        idxs = digits * xp.reshape(trans, (len(self), *[1]*(len(digits.shape)-1)))
        return xp.sum(idxs, axis=0)

    def to_digits[T: ArrayLike](self, idxs: T) -> T:
        """
        Convert dimension indices with shape (...) to digit indices with shape (len(Dimension), ...).
        """
        xp = namespace_of_arrays(idxs)
        int_type = get_index_dtype(xp)
        self._check_dtype(int_type, idxs)
        digits = xp.zeros((len(self), *idxs.shape),
                          dtype=int_type,
                          device=device(idxs))
        idxs = deepcopy(idxs)
        for i, digit in enumerate(reversed(self)):
            digits[(len(self)-i-1, ...)] = idxs % digit.base
            idxs = idxs // digit.base
        return digits

    def _check_dtype(self, index_dtype: DType, inp: ArrayLike) -> None:
        if inp.dtype != index_dtype:
            raise ValueError(f"Input should have dtype={index_dtype}")

    #-------------------------------------------------------------------------
    #some magic

    def __str__(self) -> str:
        return f"Dimension({str([d.base for d in self])})"

    def __hash__(self) -> int:
        return hash((*self[:],))

    def __eq__(self, other) -> bool:
        return isinstance(other, Dimension)\
           and len(self) == len(other)\
           and all(digits_similar(d1, d2) for d1, d2 in zip(self, other))
_idf = 0
def unique_identifier() -> int:
    global _idf
    _idf += 1
    return _idf

def prime_factorization(num: int) -> Sequence[int]:
    facs = []
    i = 2
    while i * i <= num:
        while num % i == 0:
            facs.append(i)
            num //= i
        i += 1
    if num > 1:
        facs.append(num)
    return facs

Dimensions = Sequence[Dimension]

def dimensions_similar(dim1: Dimension, dim2: Dimension) -> bool:
    return all(digits_similar(d1, d2) for d1, d2 in zip(dim1, dim2))



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Literal, Optional
from dataclasses import dataclass

from .backend import ArrayLike
from .direction import Direction
from .trainshape import TrainShape
from .matrixdecomposition import MatrixDecomposition
from .svdecomposition import SVDecomposition
from .sweepingstrategy import SweepingStrategy
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .einsumequation import EinsumEquation
from .contractor import OptimizeKind, DEFAULT_OPTIMIZER
from .utils import get_shapes

from .fullcontractor import FullContractor
from .exactcontractor import ExactContractor
from .decompositioncontractor import DecompositionContractor
from .variationalcontractor import VariationalContractor

@dataclass(frozen=True, init=False)
class EinsumExpression[T: ArrayLike]:

    equation: str
    optimizer: OptimizeKind
    method: Literal["exact", "decomposition", "variational"]
    decomposition: Optional[MatrixDecomposition]
    strategy: Optional[SweepingStrategy]
    result_shape: Optional[TrainShape]
    _expr: FullContractor | ExactContractor | DecompositionContractor | tuple[DecompositionContractor, VariationalContractor]

    def __init__(
                self,
                equation: str,
                *operands: TrainShape | TrainBase[T],
                method: Literal["exact", "decomposition", "variational"] = "decomposition",
                decomposition: Optional[MatrixDecomposition] = SVDecomposition(max_rank=25, cutoff=1e-12),
                strategy: Optional[SweepingStrategy] = SweepingStrategy(ncores=2, mode="connected", nsweeps=1),
                optimizer: OptimizeKind = DEFAULT_OPTIMIZER,
                result_shape: Optional[TrainShape] = None) -> None:

        object.__setattr__(self, "equation", equation)
        object.__setattr__(self, "optimizer", optimizer)
        object.__setattr__(self, "method", method)
        object.__setattr__(self, "decomposition", decomposition)
        object.__setattr__(self, "strategy", strategy)
        object.__setattr__(self, "result_shape", result_shape)

        eq = EinsumEquation(equation, *get_shapes(*operands))
        contr = EinsumContraction(eq, result=result_shape)

        if contr.result_shape is None:
            full_expr = FullContractor(contr, optimizer=optimizer)
            full_expr.calc_expressions(*operands)
            object.__setattr__(self, "_expr", full_expr)
            return
        elif method == "exact":
            exact_expr = ExactContractor(contr, optimizer=optimizer)
            exact_expr.calc_expressions(*operands)
            object.__setattr__(self, "_expr", exact_expr)
            return

        if decomposition is None or strategy is None:
            raise ValueError(f"Method '{method}' requires a MatrixDecomposition "\
                             f"instance and a SweepingStrategy instance.")

        if method == "decomposition" or method == "variational":
            decomp_expr = DecompositionContractor(contr,
                                                  optimizer=optimizer,
                                                  decomposition=decomposition,
                                                  strategy=strategy,
                                                  direction=Direction.TO_RIGHT)
            _, guess_shape = decomp_expr.calc_expressions(*operands)

        if method == "decomposition":
            object.__setattr__(self, "_expr", decomp_expr)
            return
        if method == "variational":
            expr = VariationalContractor(contr,
                                         optimizer=optimizer,
                                         decomposition=decomposition,
                                         strategy=strategy)
            expr.calc_expressions(guess_shape, *operands)
            object.__setattr__(self, "_expr", (decomp_expr, expr))

    def __call__(self, *operands: TrainBase[T]) -> float | TrainBase[T]:
        if isinstance(self._expr, tuple):
            return self._expr[1](self._expr[0](*operands), *operands)
        return self._expr(*operands)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .storagetensor import StorageTensor
from .backend import Device, DType, ArrayLike, get_index_dtype, namespace_of_arrays, size, shape

class ValueSet[T: ArrayLike]:
    _data: StorageTensor
    _set: set[tuple[int, ...]]

    @property
    def data(self) -> T:
        return self._data[...]

    @property
    def ndims(self) -> int:
        return self._data.max_shape[0]

    @property
    def size(self) -> int:
        return self._data.shape[1]

    @property
    def max_size(self) -> int:
        return self._data.max_shape[1]

    @property
    def dtype(self) -> DType:
        return self._data.dtype

    @property
    def device(self) -> Device:
        return self._data.device

    def __init__(self, storage: T) -> None:
        if storage.ndim != 2:
            raise ValueError("Storage tensor has incorrect shape")
        self._data = StorageTensor(storage)
        self._data.add_to_dim(0, shape(storage)[0])
        self._set = set()

    def add(self, vals: T) -> int:
        xp = namespace_of_arrays(vals)
        info = xp.__array_namespace_info__()
        btype = info.dtypes()["bool"]

        if vals.ndim == 1:
            vals = xp.reshape(vals, (size(vals), 1))
        self._check_vals(vals)

        to_add = xp.ones(vals.shape[1:], dtype=btype)
        for j in range(size(to_add)):
            key = tuple(int(val) for val in vals[:,j])
            size_ = len(self._set)
            self._set.add(key)
            to_add[j] = len(self._set) > size_
        vals = vals[:,to_add]
        added = shape(vals)[1]
        if size(vals) == 0:
            return 0

        self._data.add_to_dim(1, added)
        self._data[:,-added:] = vals
        return added

    def _check_vals(self, vals: ArrayLike) -> None:
        xp = namespace_of_arrays(vals)
        if vals.dtype != get_index_dtype(xp):
            raise ValueError("Expect a tensor of dtype torch.int64")
        if vals.ndim != 2 or vals.shape[0] != self.ndims:
            raise ValueError(f"Expect a tensor of shape ({self.ndims}, k) for some k")

    def _check_storage(self, storage: T) -> None:
        xp = namespace_of_arrays(storage[...])
        if storage.ndim != 2:
            raise ValueError("Storage tensor has incorrect shape")
        elif storage.dtype != get_index_dtype(xp):
            raise ValueError(f"Storage tensor has incorrect dtype, expected {get_index_dtype(xp)}")



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends

class TestQftShift(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [120, 256, 280, 1024]

    def test_qftshift(self):
        vals = [1.0, 2.0, 11.0]
        for ts, size, val in product(self.trainsum, self.sizes, vals):
            xp = ts.namespace
            dim = ts.dimension(size)

            exact = xp.fft.fftfreq(size, d=val) # type: ignore
            train = ts.qftfreq(dim, val)

            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2) / xp.sum(exact**2))
            self.assertLess(diff, 1e-7)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .trainsum import TrainSum



from typing import Sequence
import unittest
from itertools import product

from trainsum import TrainSum
from trainsum.typing import UniformGrid
from utils import backends

class TestMinMax(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120,), (280,), (1024,), (120, 1024), (324, 120)]

    def get_grid(self, ts, sizes: Sequence[int], lower: float, upper: float):
        dims = [ts.dimension(size) for size in sizes]
        domains = [ts.domain(lower, upper) for _ in sizes]
        return ts.uniform_grid(dims, domains)

    def get_idxs(self, ts, grid: UniformGrid):
        xp = ts.namespace
        idxs = xp.zeros([len(grid.dims), *[dim.size() for dim in grid.dims]],
                        dtype=ts.index_type)
        for i, dim in enumerate(grid.dims):
            cut = (*(xp.newaxis,) * i, slice(None), *(xp.newaxis,) * (len(grid.dims) - i - 1))
            idxs[i] += xp.arange(dim.size(), dtype=ts.index_type)[cut]
        return idxs

    def test_gauss(self) -> None:
        for sizes, ts in product(self.sizes, self.trainsum):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -10, 10)
            idxs = self.get_idxs(ts, grid)
            coords = grid.to_coords(idxs)
            
            shape = ts.trainshape(*grid.dims)
            data = xp.exp(-xp.sum(coords**2, axis=0))
            train = ts.tensortrain(shape, data)

            res = ts.min_max(train, 16)
            min_val = xp.min(data)
            max_val = xp.max(data)

            self.assertLess(abs(min_val-res.min_val), 1e-6)
            self.assertLess(abs(max_val-res.max_val), 1e-6)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Generator, Literal, Sequence
from types import NoneType
from dataclasses import dataclass
from math import prod

from .localrange import LocalRange
from .trainshape import TrainShape

@dataclass
class SweepingStrategy:
    """
    Strategy for sweeping through a tensor train. Used heavily in the arithmetic algorithms and solvers.
    """

    min_size: NoneType | int = None
    #: Number of cores that are contracted together to create a super-core.
    ncores: int = 2
    #: Mode defines the overlap of the consecutive super-cores. In "connected" mode, the super-cores overlap by one core, while in "interleaved" mode, they overlap by ncores-1.
    mode: Literal["connected", "interleaved"] = "connected"
    #: Number of sweeps to perform. If None, the strategy will sweep indefinitely.
    nsweeps: int = 1

    def __call__(self, shape: TrainShape) -> Generator[tuple[int, LocalRange]]:
        """Generator that yields the local ranges for each sweep."""
        nsweeps = 0
        begin, end = 1, len(shape)
        while True:
            if self.ncores >= len(shape):
                yield nsweeps, LocalRange(begin=0, end=len(shape))
            else:

                if begin != 0:
                    begin = 0
                elif self.mode == "connected":
                    begin = end - 1
                else:
                    begin = 1
                for lrange in self._sweep_to_right(shape, begin):
                    begin, end = lrange.begin, lrange.end
                    yield nsweeps, lrange

                if end != len(shape):
                    end = len(shape)
                elif self.mode == "connected":
                    end = min(begin + 1, len(shape) - 1)
                else:
                    end = len(shape) - 1

                for lrange in self._sweep_to_left(shape, end):
                    begin, end = lrange.begin, lrange.end
                    yield nsweeps, lrange

            nsweeps += 1
            if self.nsweeps == nsweeps:
                break

    def right_sweep(self, shape: TrainShape) -> Sequence[LocalRange]:
        """Returns the local ranges for a single right sweep."""
        if self.ncores >= len(shape):
            return [LocalRange(begin=0, end=len(shape))]

        sweep = list(self._sweep_to_right(shape, 0))
        if sweep[-1].end != len(shape):
            sweep.append(LocalRange(begin=len(shape)-self.ncores, end=len(shape)))
        return sweep

    def left_sweep(self, shape: TrainShape) -> Sequence[LocalRange]:
        """Returns the local ranges for a single left sweep."""
        if self.ncores >= len(shape):
            return [LocalRange(begin=0, end=len(shape))]

        sweep = list(self._sweep_to_left(shape, len(shape)))
        if sweep[-1].begin != 0:
            sweep.append(LocalRange(begin=0, end=self.ncores))
        return sweep

    def _sweep_to_right(self, shape: TrainShape, idx: int) -> Generator[LocalRange]:
        if self.min_size is not None and prod(dim.size() for dim in shape.dims) < 0:
            raise ValueError("min_size is too large for the given shape.")

        end = idx
        while end < len(shape):
            if self.min_size is not None:
                while end < len(shape) and self._get_size(shape, idx, end) < self.min_size:
                    end += 1
            if end - idx < self.ncores:
                end = idx + self.ncores

            if end > len(shape):
                break

            yield LocalRange(begin=idx, end=end)
            idx = max(idx+1, end-1) if self.mode == "connected" else idx + 1

    def _sweep_to_left(self, shape: TrainShape, end: int) -> Generator[LocalRange]:
        if self.min_size is not None and prod(dim.size() for dim in shape.dims) < 0:
            raise ValueError("min_size is too large for the given shape.")

        idx = end-1
        while idx > -1:
            if self.min_size is not None:
                while idx > 0 and self._get_size(shape, idx, end) < self.min_size:
                    idx -= 1
            if end - idx < self.ncores:
                idx = end - self.ncores

            if idx < 1:
                break

            yield LocalRange(begin=idx, end=end)
            end = min(idx+1, end-1) if self.mode == "connected" else end - 1

    def _get_size(self, shape: TrainShape, begin: int, end: int) -> int:
        size = 1
        for i in range(begin, end):
            size *= prod(shape.middle(i))
        size *= shape.left_rank(begin)
        size *= shape.right_rank(end-1)
        return size

    def __repr__(self) -> str:
        return (f"SweepingStrategy(ncores={self.ncores}, mode='{self.mode}', "
                f"nsweeps={self.nsweeps}, min_size={self.min_size})")



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

"""typing has all classes used in the external API of trainsum."""

from .digit import Digit
from .dimension import Dimension
from .domain import Domain
from .uniformgrid import UniformGrid
from .trainshape import TrainShape
from .tensortrain import TensorTrain

from .sweepingstrategy import SweepingStrategy
from .svdecomposition import SVDecomposition
from .qrdecomposition import QRDecomposition

from .matrixdecomposition import MatrixDecomposition
from .matrixleastsquares import MatrixLeastSquares
from .matrixeigenvaluedecomposition import MatrixEigenvalueDecomposition
from .locallinsolver import LocalLinSolver, LocalLinSolverResult
from .localeigsolver import LocalEigSolver, LocalEigSolverResult

from .options import Options, ExactOptions, EvaluationOptions, DecompositionOptions, VariationalOptions, OptionType

from .trainsum import TrainSum, EigSolver, LinSolver, LinearMap, EinsumExpression, EvaluateExpression



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends, rand_data

class TestSVDecomposition(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120, 120), (256, 250), (120, 7), (256, 800)]

    def test_construction(self):
        for ts in self.trainsum:
            decomp = ts.qrdecomposition()
    
    def test_decompose(self):
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace
            mat = rand_data(xp, *sizes)
            decomp = ts.qrdecomposition()

            res = decomp.left(mat)
            q, r = xp.linalg.qr(mat, mode='reduced')
            self.assertTrue(xp.max(res.left - q) < 1e-10)
            self.assertTrue(xp.max(res.right - r) < 1e-10)

            res = decomp.right(mat)
            r, q = xp.linalg.qr(mat.T, mode='reduced')
            q, r = q.T, r.T
            self.assertTrue(xp.max(res.left - q) < 1e-10)
            self.assertTrue(xp.max(res.right - r) < 1e-10)

if __name__ == '__main__':
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from copy import deepcopy

from .backend import ArrayLike
from .matrixdecomposition import MatrixDecomposition
from .svdecomposition import SVDecomposition
from .trainshape import TrainShape, change_dims
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .localvector import local_vector
from .sweepingstrategy import SweepingStrategy
from .contractor import OptimizeKind
from .utils import namespace_of_trains

from .einsumequation import EinsumEquation
from .innergenerator import InnerGenerator
from .generatorcallabletype import GeneratorCallableType

class VariationalContractor:

    optimizer: OptimizeKind
    decomposition: MatrixDecomposition
    strategy: SweepingStrategy

    def __init__(self,
                 contr: EinsumContraction,
                 optimizer: OptimizeKind = "greedy",
                 decomposition: MatrixDecomposition = SVDecomposition(),
                 strategy: SweepingStrategy = SweepingStrategy()) -> None:

        if contr.result_shape is None or contr.full_result_shape is None:
            raise ValueError("Variational contractor cannot be used for full ccontractions. Use FullContractor instead.")

        self.optimizer = deepcopy(optimizer)
        self.decomposition = deepcopy(decomposition)
        self.strategy = deepcopy(strategy)
        self._inner_gen = self._get_inner_generator(contr, optimizer)

    def __call__[T: ArrayLike](
            self,
            guess: TrainBase[T],
            *operands: TrainBase[T],
            expr: bool = False) -> TrainBase[T]:
        xp = namespace_of_trains(guess, *operands)
        guess = deepcopy(guess)
        igen = self._inner_gen(guess, *operands, expr=expr)
        lvec = local_vector(guess, self.decomposition)
        data = xp.zeros(0)
        gen_type = GeneratorCallableType.FULL
        try:
            for _, lrange in self.strategy(guess.shape):
                lvec.send((lrange, data))
                data = igen.send((lrange, gen_type))
        finally:
            igen.close()
        return guess

    def calc_expressions(
            self,
            guess: TrainShape | TrainBase,
            *ops: TrainShape | TrainBase) -> None:
        guess_shape = guess.shape if isinstance(guess, TrainBase) else guess
        self._inner_gen.calc_expressions(self.strategy, guess_shape, *ops)

    def _get_inner_generator(
            self,
            contr: EinsumContraction,
            optimizer: OptimizeKind
            ) -> InnerGenerator:
        eq = f"{contr.equation.result}," + ",".join(contr.equation.operands) + "->"
        ops = [contr.result_shape, *contr.equation.shapes]
        eq = EinsumEquation(eq, *ops)
        dim_map = {}
        for in_shape, out_shape in zip(ops, eq.shapes):
            for in_dim, out_dim in zip(in_shape.dims, out_shape.dims):
                dim_map[in_dim] = out_dim
        new_op_shape_dims = [dim_map[dim] for dim in contr.operation_shape.dims]
        op_shape = change_dims(contr.operation_shape, new_op_shape_dims)
        contr = EinsumContraction(eq, op_shape)
        return InnerGenerator(contr, 0, optimizer)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Callable


from .backend import ArrayLike
from .options import get_options, OptionType
from .trainbase import TrainBase
from .utils import symbol_generator, namespace_of_trains
from .crossinterpolation import CrossInterpolation
from .evaluation import Evaluation
from .min_max import min_max
from .einsumcontraction import EinsumContraction
from .einsumequation import EinsumEquation

def transform[T: ArrayLike](train: TrainBase[T], func: Callable[[T], T]) -> TrainBase[T]:
    xp = namespace_of_trains(train)
    opts = get_options(xp, OptionType.CROSS)

    mx_res = min_max(train, 8)
    if mx_res.max_val > mx_res.min_val:
        idxs = mx_res.max_idxs
    else:
        idxs = mx_res.min_idxs

    sgen = symbol_generator()
    chars = "".join(next(sgen) for _ in range(len(train.shape.dims)))
    eq = EinsumEquation(f"{chars}->{chars}", train.shape)
    contr = EinsumContraction(eq)
    expr = Evaluation(contr, optimizer="greedy")

    func_ = lambda idxs: func(expr(idxs, train))
    cross = CrossInterpolation(solver=opts.solver, strategy=opts.strategy, eps=opts.eps)
    res = cross(xp, func_, train.shape, idxs)
    return res



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Generator, Optional, Sequence, Callable
from copy import deepcopy
from string import ascii_lowercase

from .backend import ArrayLike
from .localrange import LocalRange
from .trainbase import TrainBase
from .linearmap import LinearMap
from .localvector import local_vector
from .localeigsolver import LocalEigSolver, LocalEigSolverResult
from .matrixdecomposition import MatrixDecomposition
from .utils import namespace_of_trains

from .trainshape import TrainShape, change_dims
from .innergenerator import InnerGenerator
from .sweepingstrategy import SweepingStrategy
from .svdecomposition import SVDecomposition
from .einsumequation import EinsumEquation
from .contractor import OptimizeKind, DEFAULT_OPTIMIZER
from .generatorcallabletype import GeneratorCallableType
from .operationspace import einsum_operation_shape
from .einsumcontraction import EinsumContraction

class EigSolver[T: ArrayLike, S: LocalEigSolverResult]:

    solver: LocalEigSolver[S]
    decomposition: MatrixDecomposition
    strategy: SweepingStrategy
    optimizer: OptimizeKind
    eps: float
    _maps: Sequence[LinearMap]
    _inner_gen: InnerGenerator

    def __init__(
            self,
            solver: LocalEigSolver[S],
            *maps: LinearMap[T],
            decomposition: MatrixDecomposition = SVDecomposition(),
            strategy: SweepingStrategy = SweepingStrategy(),
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER,
            eps: float = 1e-10) -> None:

        self.solver = deepcopy(solver)
        self.decomposition = deepcopy(decomposition)
        self.strategy = deepcopy(strategy)
        self.optimizer = deepcopy(optimizer)
        self.eps = eps
        self._maps = maps
        self._inner_gen = self._orth_gen(maps[0].result)

    def __call__(
            self,
            guess: TrainBase[T],
            states: Sequence[TrainBase[T]] = [], 
            expr: bool = False,
            callback: Optional[Callable[[LocalRange, S], bool]] = None,
            ) -> TrainBase[T]:
        guess = deepcopy(guess)
        gen = self._gen(guess, states, expr, callback)
        next(gen)  # warm up

        val = 0.0
        eps = []
        for i, lrange in self.strategy(guess.shape):
            loc_res = gen.send(lrange)
            if len(eps) == i:
                continue
            if i < 2:
                val = loc_res.value
                continue
            eps.append(abs(loc_res.value - val))
            val = loc_res.value
            if eps[-1] < self.eps:
                break
        return guess

    def _gen(
            self,
            guess:
            TrainBase[T],
            states: Sequence[TrainBase[T]],
            expr: bool,
            callback: Optional[Callable[[LocalRange, S], bool]] = None,
            ) -> Generator[S, LocalRange]:
        xp = namespace_of_trains(guess)

        inner_generators = [deepcopy(self._inner_gen) for _ in range(len(states))]
        #if not expr:
        #    [lmap.calc_expressions(self.strategy, guess.shape) for lmap in self._maps]
        #    [igen.calc_expressions(self.strategy, guess.shape, state.shape) for igen, state in zip(inner_generators, states)]

        lvec = local_vector(guess, self.decomposition)
        map_gens = [lin_map(guess, expr=expr) for lin_map in self._maps]
        inner_gens = [igen(guess, state, expr=expr) for igen, state in zip(inner_generators, states)]
        gen_type = GeneratorCallableType.FULL

        loc_res = None
        data = xp.empty(0)
        try:
            while True:
                lrange = yield loc_res # type: ignore
                vec = lvec.send((lrange, data))
                funcs = [gen.send((lrange, gen_type)) for gen in map_gens]
                func = lambda x: sum((f(x) for f in funcs[1:]), start=funcs[0](x))
                lstates = [gen.send((lrange, gen_type)) for gen in inner_gens]
                loc_res = self.solver(func, vec, lstates) # type: ignore
                if callback is not None:
                    if callback(lrange, loc_res):
                        break
                data = loc_res.array
        finally:
            for gen in map_gens:
                gen.close()
            lvec.close()

    def _orth_gen(self, state: TrainShape) -> InnerGenerator:
        dims = ascii_lowercase[:len(state.dims)]
        eq = EinsumEquation(f"{dims},{dims}->{dims}", state, state)
        op_shape = einsum_operation_shape(eq)
        eq_ = EinsumEquation(f"{dims},{dims}->", state, state)
        dim_map = {}
        for in_shape, out_shape in zip(eq.shapes, eq_.shapes):
            for in_dim, out_dim in zip(in_shape.dims, out_shape.dims):
                dim_map[in_dim] = out_dim
        new_op_shape_dims = [dim_map[dim] for dim in op_shape.dims]
        contr = EinsumContraction(eq_, op_shape=change_dims(op_shape, new_op_shape_dims))
        return InnerGenerator(contr, 0, self.optimizer)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from dataclasses import dataclass

from .backend import ArrayNamespace, Device, DType
from .trainshape import TrainShape
from .trainbase import TrainBase
from .utils import namespace_of_trains, get_device_dtype

@dataclass(frozen=True, init=False)
class ContractorInput:

    idx_map: dict[int, int]
    shapes: Sequence[TrainShape]
    _device: None | Device
    _dtype: None | DType
    _xp: None | ArrayNamespace

    def __init__(self, *ops: TrainShape | TrainBase):
        idx = 0
        shapes = []
        idx_map = {}
        bases = []
        for i, op in enumerate(ops):
            if isinstance(op, TrainShape):
                shapes.append(op)
                idx_map[i] = idx
                idx += 1
            elif isinstance(op, TrainBase):
                bases.append(op)
        if len(bases) > 0:
            xp = namespace_of_trains(*bases)
            device, dtype = get_device_dtype(bases)
            object.__setattr__(self, "_device", device)
            object.__setattr__(self, "_dtype", dtype)
            object.__setattr__(self, "_xp", xp)
        else:
            object.__setattr__(self, "_device", None)
            object.__setattr__(self, "_dtype", None)
            object.__setattr__(self, "_xp", None)
        object.__setattr__(self, "shapes", shapes)
        object.__setattr__(self, "idx_map", idx_map)

    def check_operands(self, *ops: TrainBase) -> None:
        if len(self.shapes) != len(ops):
            raise ValueError("Number of operand shapes do not match contraction")
        for ref_op, op in zip(self.shapes, (op.shape for op in ops)):
            if not ref_op == op:
                raise ValueError("Operand shapes do not match contraction")
        if self._device is None and self._dtype is None:
            ref_device = ops[0].device
            ref_dtype = ops[0].dtype
        else:
            ref_device = self._device
            ref_dtype = self._dtype
        if not all(op.device == ref_device for op in ops):
            raise ValueError("All operand devices must match")
        if not all(op.dtype == ref_dtype for op in ops):
            raise ValueError("All operand dtypes must match")

    def infos(self, *ops: TrainBase) -> tuple[ArrayNamespace, Device, DType]:
        if self._xp is None or self._device is None or self._dtype is None:
            xp = namespace_of_trains(*ops)
            device, dtype = get_device_dtype(ops)
            return xp, device, dtype
        return self._xp, self._device, self._dtype



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from copy import deepcopy

from .backend import ArrayNamespace, ArrayLike
from .matrixdecomposition import MatrixDecomposition
from .dimension import Dimension
from .trainshape import TrainShape
from .trainbase import TrainBase

from .sweepingstrategy import SweepingStrategy
from .svdecomposition import SVDecomposition
from .einsumexpression import EinsumExpression
from .domain import Domain
from .uniformgrid import UniformGrid
from .polyval import polyval
from .conj import conj
from .utils import namespace_of_trains

def qft[T: ArrayLike](
        xp: ArrayNamespace[T],
        dim: Dimension,
        decomp: MatrixDecomposition = SVDecomposition(max_rank=16)
        ) -> TrainBase:

    layers = [fft_layer(xp, dim, i) for i in range(len(dim))]

    strat = SweepingStrategy(ncores=2)
    ref_shape = deepcopy(layers[0].shape)
    ref_shape.ranks = 16
    layer_shape = deepcopy(layers[0].shape)
    layer_shape.ranks = 2
    expr = EinsumExpression(
        "ab,ab->ab", ref_shape, layer_shape,
        method="decomposition",
        decomposition=decomp,
        strategy=strat)

    res = layers[0]
    for l in layers[1:]:
        res = expr(res, l)
        if not isinstance(res, TrainBase):
            raise ValueError("Expected a TrainBase as result here.")
    res.data[0][...] /= xp.sqrt(xp.asarray(dim.size()))
    return res

def iqft[T: ArrayLike](
        xp: ArrayNamespace[T],
        dim: Dimension,
        decomp: MatrixDecomposition = SVDecomposition(max_rank=16)
        ) -> TrainBase:
    op = conj(qft(xp, dim, decomp))
    shape = TrainShape(op.shape.dims[::-1], op.shape.digits)
    return TrainBase(shape, op.data, copy_data=False)

def fft_layer[T: ArrayLike](
        xp: ArrayNamespace[T],
        dim: Dimension,
        idx: int
        ) -> TrainBase[T]:
    info = xp.__array_namespace_info__()
    ctype = info.dtypes()["complex128"]
    factor = -2*xp.pi*1.0j/dim.size()
    ref_xi = dim[idx].factor
    rank = dim[idx].base

    dims = (Dimension([d.base for d in dim]),
            Dimension([d.base for d in dim[::-1]]))

    data = []
    for i, (row_digit, col_digit) in enumerate(zip(dims[0], reversed(dims[1]))):
        tmp = xp.zeros((rank, row_digit.base, col_digit.base, rank), dtype=ctype)
        if i == idx:
            for j in range(row_digit.base):
                for k in range(col_digit.base):
                    row_fac = row_digit.factor * j
                    col_fac = col_digit.factor * k
                    tmp[j,j,k,j] = xp.exp(xp.asarray(factor*row_fac*col_fac))
        else:
            for j in range(row_digit.base):
                for k in range(col_digit.base):
                    for l in range(rank):
                        row_fac = ref_xi * l
                        col_fac = col_digit.factor * k
                        tmp[l,j,k,l] = xp.exp(xp.asarray(factor*row_fac*col_fac))
        data.append(tmp)
    tmp = xp.ones(rank, dtype=ctype)
    data[0] = xp.tensordot(data[0], tmp, axes=([0],[0]))
    data[0] = xp.expand_dims(data[0], axis=0)
    data[-1] = xp.tensordot(tmp, data[-1], axes=([0],[3]))
    data[-1] = xp.expand_dims(data[-1], axis=3)

    shape = TrainShape(dims[::-1], [digits for digits in zip(dims[0], reversed(dims[1]))])
    return TrainBase(shape, data)

def qftshift[T: ArrayLike](train: TrainBase[T], axis: int=0) -> TrainBase[T]:
    if 0 < axis < len(train.shape.dims):
        raise ValueError("Axis out of range.")
    dim = train.shape.dims[axis]
    xp = namespace_of_trains(train)
    if dim[0].base != 2:
        raise NotImplementedError("QFT shift is only implemented for dimensions with leading bit 2.")
    res = deepcopy(train)
    for i, digits in enumerate(res.shape.digits):
        if dim[0] in digits:
            idx = digits.index(dim[0])
            res.data[i][...] = xp.flip(res.data[i], axis=idx+1)
    return res

def iqftshift[T: ArrayLike](train: TrainBase[T], axis: int=0) -> TrainBase[T]:
    return qftshift(train, axis)

def qftfreq[T: ArrayLike](
        xp: ArrayNamespace[T],
        dim: Dimension,
        d: float
        ) -> TrainBase[T]:
    if dim.size() % 2 == 0: # even
        domain = Domain(-dim.size()/2, dim.size()/2 - 1)
    else: # odd
        domain = Domain(-(dim.size()-1) / 2, -(dim.size()-1) / 2)
    grid = UniformGrid(dim, domain)
    freqs = polyval(xp, grid, [1.0, 0.0], 0.0)
    freqs = qftshift(freqs)
    freqs.data[0][...] /= dim.size()*d
    return freqs



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from copy import deepcopy

from .backend import ArrayLike
from .direction import Direction
from .localrange import LocalRange
from .normalization import Normalization
from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction, get_symbol_generator
from .matrixdecomposition import MatrixDecomposition
from .tensordecomposition import TensorDecomposition
from .svdecomposition import SVDecomposition
from .sweepingstrategy import SweepingStrategy
from .utils import check_operand_shapes, get_shapes, shape_map
from .contractor import ArrayContractor, OptimizeKind
from .contractorinput import ContractorInput
from .localcontraction import fuse_local_contractions, LocalContraction
from .contains import contains

class DecompositionContractor:

    optimizer: OptimizeKind
    decomposition: MatrixDecomposition
    strategy: SweepingStrategy
    direction: Direction
    _contr: EinsumContraction
    _inp: None | ContractorInput = None
    _tmp_str: str
    _cmap: Sequence[int]
    _to_right_exprs: dict[tuple[int, LocalRange],
                          tuple[LocalContraction, ArrayContractor]]
    _to_right_tranges: list[LocalRange]
    _to_left_exprs: dict[tuple[LocalRange, int],
                         tuple[LocalContraction, ArrayContractor]]
    _to_left_tranges: list[LocalRange]

    def __init__(self,
                 contr: EinsumContraction,
                 optimizer: OptimizeKind = "greedy",
                 decomposition: MatrixDecomposition = SVDecomposition(),
                 strategy: SweepingStrategy = SweepingStrategy(),
                 direction: Direction = Direction.TO_RIGHT) -> None:

        if contr.result_shape is None or contr.full_result_shape is None:
            raise ValueError("DecompositionContractor cannot be used for full ccontractions. Use FullContractor instead.")

        self.optimizer = deepcopy(optimizer)
        self.decomposition = deepcopy(decomposition)
        self.strategy = deepcopy(strategy)
        self.direction = direction

        self._contr = deepcopy(contr)
        res = contains(contr.full_result_shape, contr.result_shape)
        self._cmap = res.core_idxs
        self._tmp_str = next(get_symbol_generator(contr))

        self._inp = None
        self._to_right_exprs = {}
        self._to_right_tranges = []
        self._to_left_exprs = {}
        self._to_left_tranges = []

    def __call__[T: ArrayLike](self,
            *ops: TrainBase[T],
            expr: bool = False) -> TrainBase[T]:
        shapes = get_shapes(*ops)
        if expr or self._inp is None:
            self.calc_expressions(*shapes)
        else:
            self._inp.check_operands(*ops)

        if self.direction == Direction.TO_RIGHT:
            [op.normalize(0) for op in ops]
            return self._contract_to_right(self.decomposition, *ops)
        elif self.direction == Direction.TO_LEFT:
            [op.normalize(-1) for op in ops]
            return self._contract_to_left(self.decomposition, *ops)
        raise ValueError("Direction must be either 'to_left' or 'to_right'.")

    def calc_expressions(self, *ops: TrainShape | TrainBase) -> tuple[TrainShape, TrainShape]:
        check_operand_shapes(self._contr.operand_shapes, get_shapes(*ops))
        self._inp = ContractorInput(*ops)

        right_ranges = self._get_ranges(Direction.TO_RIGHT)
        right_shape = self._to_right_guess(self.decomposition, right_ranges, *ops)
        self._to_right_expressions(right_ranges, right_shape, *ops)

        left_ranges = self._get_ranges(Direction.TO_LEFT)
        left_shape = self._to_left_guess(self.decomposition, left_ranges, *ops)
        self._to_left_expressions(left_ranges, left_shape, *ops)
        return left_shape, right_shape

    # ------------------------------------------------------------------------
    # Contraction generators

    def _contract_to_right[T: ArrayLike](
            self,
            decomp: MatrixDecomposition,
            *ops: TrainBase[T]) -> TrainBase[T]:
        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        xp, device, dtype = self._inp.infos(*ops)

        shape = self._contr.result_shape
        data = []
        norm = []
        if shape is None:
            raise ValueError("Result shape cannot be none here.")
        tn_decomp = TensorDecomposition(decomp)

        lidx = 0
        idx = 0
        tmp = xp.ones([1, *[1]*len(self._contr[0].result.left)],
                      device=device, dtype=dtype)
        for i, trange in enumerate(self._to_right_tranges):
            lcontr, expr = self._to_right_exprs[(idx, trange)]
            tns = lcontr.get_data(*ops, idx_map=self._inp.idx_map)
            tmp = expr(tmp, *tns)
            idx = trange.end
            
            ref_idx = self._cmap[-1]+1
            if i < len(self._to_right_tranges)-1:
                ref_idx = self._to_right_tranges[i+1].begin

            while lidx < len(shape)-1 and self._cmap[lidx] < ref_idx:
                split = 1+len(shape.middle(lidx))
                #left, tmp = tn_decomp.left(tmp, split)
                res = tn_decomp.left(tmp, split)
                left, tmp = res.left, res.right
                data.append(left)
                norm.append(Normalization.LEFT)
                lidx += 1
        data.append(tmp)
        norm.append(Normalization.NONE)

        return TrainBase(shape, data, norm, copy_data=False)

    def _contract_to_left[T: ArrayLike](
            self,
            decomp: MatrixDecomposition,
            *ops: TrainBase[T]) -> TrainBase[T]:
        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        xp, device, dtype = self._inp.infos(*ops)

        shape = self._contr.result_shape
        data = []
        norm = []
        if shape is None:
            raise ValueError("Result shape cannot be none here.")
        tn_decomp = TensorDecomposition(decomp)

        lidx = len(shape)-1
        idx = len(self._contr)
        tmp = xp.ones([*[1]*len(self._contr[-1].result.right), 1],
                      device=device, dtype=dtype)
        for i, trange in enumerate(self._to_left_tranges):
            lcontr, expr = self._to_left_exprs[(trange, idx)]
            tns = lcontr.get_data(*ops, idx_map=self._inp.idx_map)
            tmp = expr(*tns, tmp)
            idx = trange.begin
            
            ref_idx = self._cmap[0]-1
            if i < len(self._to_left_tranges)-1:
                ref_idx = self._to_left_tranges[i+1].end-1

            while lidx > 0 and self._cmap[lidx] > ref_idx:
                split = len(tmp.shape)-len(shape.middle(lidx))-1
                #tmp, right = tn_decomp.right(tmp, split)
                res = tn_decomp.right(tmp, split)
                tmp, right = res.left, res.right
                data.append(right)
                norm.append(Normalization.RIGHT)
                lidx -= 1

        data.append(tmp)
        data.reverse()
        norm.append(Normalization.NONE)
        norm.reverse()
        return TrainBase(shape, data, norm, copy_data=False)

    # ------------------------------------------------------------------------
    # Expression builders

    def _to_right_expressions(self,
                              tranges: Sequence[LocalRange],
                              guess: TrainShape,
                              *ops: TrainShape | TrainBase) -> TrainShape:
        self._to_right_exprs.clear()
        self._to_right_tranges.clear()

        self._to_right_expression(0, tranges[0], 1, *ops)
        self._to_right_tranges.append(tranges[0])
        for i in range(1, len(tranges)):
            end_idx = tranges[i-1].end
            left_size = guess.right_rank(end_idx-1)
            if (end_idx, tranges[i]) in self._to_right_exprs:
                continue
            elif end_idx >= tranges[i].end:
                continue
            self._to_right_expression(end_idx, tranges[i], left_size, *ops)
            self._to_right_tranges.append(tranges[i])
        return guess

    def _to_right_guess(
            self,
            decomp: MatrixDecomposition,
            tranges: Sequence[LocalRange],
            *ops: TrainShape | TrainBase) -> TrainShape:
        res_shape = self._contr.result_shape
        if res_shape is None:
            raise ValueError("Result shape cannot be none here.")
        tn_decomp = TensorDecomposition(decomp)
        ranks = [1]

        idx = 0
        for i, trange in enumerate(tranges):
            lcontr = fuse_local_contractions(self._contr[trange.begin:trange.end])
            smap = shape_map(ops, lcontr)
            smap[self._tmp_str] = ranks[-1]
            result = f"{self._tmp_str}{lcontr.result.middle}{lcontr.result.right}"
            tmp = [smap[char] for char in result]
            
            ref_idx = self._cmap[-1]+1
            if i < len(tranges)-1:
                ref_idx = tranges[i+1].begin

            while idx < len(res_shape)-1 and self._cmap[idx] < ref_idx:
                split = 1+len(res_shape.middle(idx))
                left, tmp = tn_decomp.left_shape(tuple(tmp), split)
                ranks.append(left[-1])
                idx += 1
        return TrainShape(res_shape.dims, res_shape.digits, ranks[1:])

    def _to_right_expression(self,
                             prev_tend: int,
                             cur_trange: LocalRange,
                             left_size: int,
                             *ops: TrainBase | TrainShape
                             ) -> None:
        if prev_tend < cur_trange.begin:
            raise ValueError("End of previous range must be at least the begin of the current range.")
        elif prev_tend >= cur_trange.end:
            raise ValueError("End of previous range must be less than the end of the current range.")

        res_contr = fuse_local_contractions(self._contr[cur_trange.begin:cur_trange.end])
        res_str = f"{self._tmp_str}{res_contr.result.middle}{res_contr.result.right}"

        if prev_tend == cur_trange.begin:
            tmp_str = f"{self._tmp_str}{res_contr.result.left}"
            contr = res_contr
        else:
            tmp_contr = fuse_local_contractions(self._contr[cur_trange.begin:prev_tend])
            tmp_str = f"{self._tmp_str}{tmp_contr.result.middle}{tmp_contr.result.right}"
            contr = fuse_local_contractions(self._contr[prev_tend:cur_trange.end])
        eq = f"{tmp_str}," + ",".join(str(op) for op in contr.operands) + f"->{res_str}"

        smap = shape_map(ops, *self._contr[cur_trange.begin:cur_trange.end])
        smap[self._tmp_str] = left_size
        tmp_shape = [smap[char] for char in tmp_str]

        _, tns = contr.get_constants(*ops)
        expr = ArrayContractor(eq, tmp_shape, *tns, optimizer=self.optimizer)
        self._to_right_exprs[(prev_tend, cur_trange)] = contr, expr

    def _to_left_expressions(self,
                             tranges: Sequence[LocalRange],
                             guess: TrainShape,
                             *ops: TrainShape | TrainBase) -> None:
        self._to_left_exprs.clear()
        self._to_left_tranges.clear()

        self._to_left_expression(tranges[-1], tranges[-1].end, guess.left_rank(-1), *ops)
        self._to_left_tranges.append(tranges[-1])
        for i in range(len(tranges)-2, -1, -1):
            begin_idx = tranges[i+1].begin
            left_size = guess.left_rank(begin_idx)
            if (tranges[i], begin_idx) in self._to_left_exprs:
                continue
            elif begin_idx <= tranges[i].begin:
                continue
            self._to_left_expression(tranges[i], begin_idx, left_size, *ops)
            self._to_left_tranges.append(tranges[i])

    def _to_left_guess(
            self,
            decomp: MatrixDecomposition,
            tranges: Sequence[LocalRange],
            *ops: TrainShape | TrainBase) -> TrainShape:
        res_shape = self._contr.result_shape
        if res_shape is None:
            raise ValueError("Result shape cannot be none here.")
        tn_decomp = TensorDecomposition(decomp)
        ranks = [1]

        idx = len(res_shape)-1
        for i, trange in enumerate(reversed(tranges)):
            lcontr = fuse_local_contractions(self._contr[trange.begin:trange.end])
            smap = shape_map(ops, lcontr)
            smap[self._tmp_str] = ranks[-1]
            result = f"{lcontr.result.left}{lcontr.result.middle}{self._tmp_str}"
            tmp = [smap[char] for char in result]
            
            ref_idx = self._cmap[0]-1
            if i < len(tranges)-1:
                ref_idx = tranges[-i-2].end-1

            while idx > 0 and self._cmap[idx] > ref_idx:
                split = len(tmp)-len(res_shape.middle(idx))-1
                tmp, right = tn_decomp.right_shape(tuple(tmp), split)
                ranks.append(right[0])
                idx -= 1
        ranks.reverse()
        return TrainShape(res_shape.dims, res_shape.digits, ranks[1:])

    def _to_left_expression(self,
                            cur_trange: LocalRange,
                            prev_tbegin: int,
                            left_size: int,
                            *ops: TrainBase | TrainShape
                            ) -> None:
        if prev_tbegin > cur_trange.end:
            raise ValueError("Begin of the previous range must be at least the end of the current range.")
        elif prev_tbegin <= cur_trange.begin:
            raise ValueError("Begin of the previous range must be greater than the begin of the current range.")

        res_contr = fuse_local_contractions(self._contr[cur_trange.begin:cur_trange.end])
        res_str = f"{res_contr.result.left}{res_contr.result.middle}{self._tmp_str}"

        if prev_tbegin == cur_trange.end:
            tmp_str = f"{res_contr.result.right}{self._tmp_str}"
            contr = res_contr
        else:
            tmp_contr = fuse_local_contractions(self._contr[prev_tbegin:cur_trange.end])
            tmp_str = f"{tmp_contr.result.left}{tmp_contr.result.middle}{self._tmp_str}"
            contr = fuse_local_contractions(self._contr[cur_trange.begin:prev_tbegin])
        eq = ",".join(str(op) for op in contr.operands) + f",{tmp_str}" + f"->{res_str}"

        smap = shape_map(ops, *self._contr[cur_trange.begin:cur_trange.end])
        smap[self._tmp_str] = left_size
        tmp_shape = [smap[char] for char in tmp_str]

        _, tns = contr.get_constants(*ops)
        expr = ArrayContractor(eq, *tns, tmp_shape, optimizer=self.optimizer)
        self._to_left_exprs[(cur_trange, prev_tbegin)] = contr, expr


    def _transform_range(self, lrange: LocalRange) -> LocalRange:
        begin = self._cmap[lrange.begin]
        end = self._cmap[lrange.end-1]
        return LocalRange(begin=begin, end=end+1)

    def _get_ranges(self, direction: Direction) -> Sequence[LocalRange]:
        if self._contr.result_shape is None:
            raise RuntimeError("Result shape cannot be none here.")
        if direction == Direction.TO_RIGHT:
            lranges = self.strategy.right_sweep(self._contr.result_shape)
        elif direction == Direction.TO_LEFT:
            lranges = self.strategy.left_sweep(self._contr.result_shape)
        tranges = [self._transform_range(lrange) for lrange in lranges]
        tranges = sorted(tranges)
        tranges = list(dict.fromkeys(tranges))
        if tranges[0].begin != 0 or tranges[-1].end != len(self._contr):
            raise RuntimeError("Provided local ranges do not cover the full contraction.")
        return tranges



from typing import Any
import unittest
from math import prod
from itertools import product

from trainsum import TrainSum
from trainsum.digit import Digit
from trainsum.backend import get_index_dtype
from utils import prime_factorization, backends

class TestDimension(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [120, 256]

    def digits(self, idf: int, size: int) -> list[Digit]:
        bases = prime_factorization(size)
        factors = [prod(bases[i+1:]) for i in range(len(bases))]
        return [Digit(idf, i, base, factor) for i, (base, factor) in enumerate(zip(bases, factors))]

    def digit_tensor(self, xp: Any, size: int) -> Any:
        int_type = get_index_dtype(xp)
        bases = prime_factorization(size)
        idxs = xp.arange(size, dtype=int_type)
        digits = xp.zeros((len(bases), *idxs.shape), dtype=int_type)
        for i, base in enumerate(reversed(bases)):
            digits[len(bases)-i-1,...] = idxs % base
            idxs //= base
        return digits

    def index_tensor(self, xp: Any, size: int) -> Any:
        int_type = get_index_dtype(xp)
        return xp.arange(0, size, dtype=int_type)

    def test_construction(self):
        for ts, size in product(self.trainsum, self.sizes):
            dim = ts.dimension(size)
            ref_digits = self.digits(dim.idf, size)
            self.assertEqual(len(dim), len(ref_digits))
            self.assertTrue(all(d1==d2 for d1, d2 in zip(dim, ref_digits)))

        for ts in self.trainsum:
            self.assertRaises(ValueError, ts.dimension, 0)
            self.assertRaises(ValueError, ts.dimension, -1)
            self.assertRaises(ValueError, ts.dimension, [])
            self.assertRaises(ValueError, ts.dimension, [-1, 2])
            self.assertRaises(ValueError, ts.dimension, [0, 2])

    def test_unique_idf(self):
        for ts in self.trainsum:
            dim1 = ts.dimension(10)
            dim2 = ts.dimension(10)
            self.assertNotEqual(dim1.idf, dim2.idf)

    def test_eq(self):
        for ts in self.trainsum:
            dims = [ts.dimension(size) for size in self.sizes]
            for i in range(len(dims)):
                self.assertEqual(dims[i], dims[i])
                for j in range(i+1, len(dims)):
                    self.assertNotEqual(dims[i], dims[j])

    def test_to_idxs(self):
        for (xp, ts), size in product(zip(backends, self.trainsum), self.sizes):
            ref_idxs = self.index_tensor(xp, size)

            dim = ts.dimension(size)
            digits = self.digit_tensor(xp, size)
            idxs = dim.to_idxs(digits)
            self.assertTrue(xp.all(idxs == ref_idxs))

    def test_to_digits(self):
        for (xp, ts), size in product(zip(backends, self.trainsum), self.sizes):
            ref_digits = self.digit_tensor(xp, size)

            dim = ts.dimension(size)
            idxs = self.index_tensor(xp, size)
            digits = dim.to_digits(idxs)
            self.assertTrue(xp.all(digits == ref_digits))
    
if __name__ == '__main__':
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Literal

from .backend import ArrayLike, ArrayNamespace, get_index_dtype
from .dimension import Dimension
from .trainshape import TrainShape
from .trainbase import TrainBase

def toeplitz[T: ArrayLike](
        xp: ArrayNamespace[T],
        dim: Dimension,
        mode: Literal["full", "lower", "upper", "circular"]
        ) -> TrainBase:
    if mode == "full":
        dim = Dimension([2, *[d.base for d in dim]])

    out = []
    for digit in dim:
        data = xp.zeros((2, digit.base, digit.base, digit.base, 2))
        for j in range(digit.base):
            data[:,j,...] = shift_core(xp, j, digit.base)
        out.append(data)
    out[-1] = xp.reshape(out[-1][...,0], [*out[-1].shape[:-1], 1])
    if mode == "full":
        out[0] = xp.zeros([1, 2, 2])
        out[0][0,:,0] = xp.asarray([0.0, 1.0])
        out[0][0,:,1] = xp.asarray([1.0, 0.0])
        dims = [Dimension([d.base for d in dim]),
                Dimension([d.base for d in dim[1:]]),
                Dimension([d.base for d in dim[1:]])]
        shape = TrainShape(dims, [(dims[0][0],)] + [d for d in zip(dims[0][1:], dims[1], dims[2])])
    else:
        w = {"lower": xp.asarray([1.0, 0.0]),
             "upper": xp.asarray([0.0, 1.0]),
             "circular": xp.asarray([1.0, 1.0])
             }[mode]
        tmp = xp.tensordot(w, out[0], axes=([0], [0]))
        out[0] = xp.reshape(tmp, [1, *tmp.shape])
        dims = [Dimension([d.base for d in dim]),
                Dimension([d.base for d in dim]),
                Dimension([d.base for d in dim])]
        shape = TrainShape(dims, [d for d in zip(dims[0], dims[1], dims[2])])
    return TrainBase(shape, out)

def shift[T: ArrayLike](
        xp: ArrayNamespace[T],
        dim: Dimension,
        shift: int,
        circular: bool = False
        ) -> TrainBase:
    dims = Dimension([d.base for d in dim]), Dimension([d.base for d in dim])
    shape = TrainShape(dims, [d for d in zip(*dims)])

    if shift == 0:
        data = [xp.eye(digit.base).reshape(1, digit.base, digit.base, 1) for digit in dim]
        return TrainBase(shape, data)
    elif abs(shift) >= dim.size():
        raise ValueError("Shift value must be less than dimension size")

    if shift < 0:
        shift = abs(shift)
        w = xp.asarray([1.0, 0.0]) if not circular else xp.asarray([1.0, 1.0])
    else:
        shift = dim.size() - shift
        w = xp.asarray([0.0, 1.0]) if not circular else xp.asarray([1.0, 1.0])

    idxs = xp.asarray([[shift]], dtype=get_index_dtype(xp))
    digits = dim.to_digits(idxs)[:,0]

    data = []
    for num, digit in zip(digits, dim):
        data.append(shift_core(xp, int(num[0]), digit.base))
    tmp = xp.tensordot(w, data[0], axes=([0], [0]))
    data[0] = xp.reshape(tmp, (1, *tmp.shape))
    data[-1] = data[-1][...,0:1]
    return TrainBase(shape, data)

def shift_core[T: ArrayLike](
        xp: ArrayNamespace[T],
        digit: int,
        base: int
        ) -> T:
    core = xp.zeros((2, base, base, 2))
    for c_in in range(2):
        for j in range(base):
            s = j + digit + c_in
            i = s % base
            c_out = s // base  # 0 or 1
            #core = set_data(core, (c_out, i, j, c_in), 1.0)
            core[c_out,i,j,c_in] = 1.0
    return core



from typing import Sequence
import unittest
from itertools import product

from trainsum import TrainSum
from trainsum.typing import UniformGrid
from utils import backends

class TestEigsolver(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(1024,)]

    def get_grid(self, ts, sizes: Sequence[int], lower: float, upper: float):
        dims = [ts.dimension(size) for size in sizes]
        domains = [ts.domain(lower, upper) for _ in sizes]
        return ts.uniform_grid(dims, domains)

    def get_idxs(self, ts, grid: UniformGrid):
        xp = ts.namespace
        idxs = xp.zeros([len(grid.dims), *[dim.size() for dim in grid.dims]],
                        dtype=ts.index_type)
        for i, dim in enumerate(grid.dims):
            cut = (*(xp.newaxis,) * i, slice(None), *(xp.newaxis,) * (len(grid.dims) - i - 1))
            idxs[i] += xp.arange(dim.size(), dtype=ts.index_type)[cut]
        return idxs

    def test_solve(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -20.0, 20.0)
            idxs = self.get_idxs(ts, grid)
            coords = grid.to_coords(idxs)

            shape = ts.trainshape(*sizes)
            data = xp.exp(-0.001*xp.sum(coords**2, axis=0))
            data /= xp.sum(data**2)
            guess = ts.tensortrain(shape, data)
            pot = ts.polyval(grid, [1.0, 0.0, 0.0], 0.0)
            with ts.exact():
                lap_op = -2*ts.shift(grid.dims[0], 0)
                lap_op += ts.shift(grid.dims[0], 1)
                lap_op += ts.shift(grid.dims[0], -1)
                lap_op *= -0.5/grid.spacings[0]

            decomp = ts.svdecomposition(max_rank=15, cutoff=1e-15)
            strat = ts.sweeping_strategy(ncores=2, nsweeps=20)
            solver = ts.lanczos(subspace=3, nsteps=1, eps=1e-8)

            lap_map = ts.linear_map("ij,j->i", lap_op, guess.shape)
            pot_map = ts.linear_map("i,i->i", pot, guess.shape)

            eigsolver = ts.eigsolver(lap_map, pot_map,
                                     solver=solver,
                                     decomposition=decomp,
                                     strategy=strat)
            eigvals = []
            def call(lrange, res):
                eigvals.append(res.value)
                return False
            guess = eigsolver(guess, callback=call)

            mat = lap_op.to_tensor()
            mat += xp.eye(mat.shape[0]) * pot.to_tensor()
            exact = xp.linalg.eigh(mat)[0][0]
            self.assertTrue(abs(eigvals[-1] - exact) < 1e-6)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from types import NoneType
from copy import deepcopy
from math import prod

from .backend import ArrayLike
from .trainshape import TrainShape
from .trainbase import Normalization, TrainBase
from .einsumcontraction import EinsumContraction
from .contractor import ArrayContractor, OptimizeKind
from .contractorinput import ContractorInput
from .utils import check_operand_shapes, get_shapes

class ExactContractor:

    optimizer: OptimizeKind
    _contr: EinsumContraction
    _inp: NoneType | ContractorInput = None
    _exprs: NoneType | Sequence[ArrayContractor] = None

    def __init__(self,
                 contr: EinsumContraction,
                 optimizer: OptimizeKind = "greedy") -> None:
        if contr.result_shape is None:
            raise ValueError("ExactContractor requires a result shape.")

        self.optimizer = deepcopy(optimizer)
        self._contr = deepcopy(contr)

    def __call__[T: ArrayLike](
            self,
            *ops: TrainBase[T],
            expr: bool = False) -> TrainBase[T]:
        #get_device_dtype(ops)
        shapes = get_shapes(*ops)
        if expr or self._inp is None or self._exprs is None:
            self.calc_expressions(*shapes)
        else:
            self._inp.check_operands(*ops)
        return self._contract(*ops, expr=expr)

    def calc_expressions(self, *ops: TrainShape | TrainBase) -> None:
        check_operand_shapes(self._contr.operand_shapes, get_shapes(*ops))
        self._inp = ContractorInput(*ops)
        self._exprs = [self._expression(i, *ops) for i in range(len(self._contr))]

    def _expression(self, idx: int, *shapes: TrainShape | TrainBase) -> ArrayContractor:
        lcontr = self._contr[idx]
        eq = ",".join(str(op) for op in lcontr.operands) + f"->{lcontr.result}"
        tns = lcontr.get_shapes(*shapes)
        _, tns = lcontr.get_constants(*shapes)
        expr = ArrayContractor(eq, *tns, optimizer=self.optimizer)
        return expr

    def _contract(self, *ops: TrainBase, expr: bool = False) -> TrainBase:
        if self._contr.full_result_shape is None:
            raise ValueError("Contraction needs a result.")
        elif self._inp is None or self._exprs is None:
            raise RuntimeError("Expression or input cannot be None here.")

        xp, _, _ = self._inp.infos(*ops)

        core_data = []
        for i, contr in enumerate(self._contr):
            if expr:
                tns = contr.get_data(*ops)
                data = self._expression(i, *ops)(*tns)
            else:
                tns = contr.get_data(*ops, idx_map=self._inp.idx_map)
                data = self._exprs[i](*tns)

            dleft = len(contr.result.left)
            dright = len(contr.result.right)
            left_shape = prod(data.shape[:dleft])
            right_shape = prod(data.shape[-dright:])
            middle_shape = data.shape[dleft:-dright]
            data = xp.reshape(data, (left_shape, *middle_shape, right_shape))
            core_data.append(data)
        norm = [Normalization.NONE] * len(core_data)
        return TrainBase(self._contr.full_result_shape, core_data,
                         norm=norm, copy_data=False)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from copy import deepcopy
from string import ascii_lowercase

from .backend import ArrayLike
from .matrixdecomposition import MatrixDecomposition
from .svdecomposition import SVDecomposition
from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .localvector import local_vector
from .sweepingstrategy import SweepingStrategy
from .contractor import OptimizeKind
from .utils import namespace_of_trains, trains_match

from .einsumequation import EinsumEquation
from .innergenerator import InnerGenerator
from .generatorcallabletype import GeneratorCallableType

class VariationalAddition:

    optimizer: OptimizeKind
    decomposition: MatrixDecomposition
    strategy: SweepingStrategy

    def __init__(
            self,
            shape: TrainShape,
            decomposition: MatrixDecomposition = SVDecomposition(),
            strategy: SweepingStrategy = SweepingStrategy(),
            optimizer: OptimizeKind = "greedy") -> None:

        self.optimizer = deepcopy(optimizer)
        self.decomposition = deepcopy(decomposition)
        self.strategy = deepcopy(strategy)
        self._inner_gen = self._get_inner_generator(shape, optimizer)

    def _get_inner_generator(
            self,
            shape: TrainShape,
            optimizer: OptimizeKind
            ) -> InnerGenerator:
        dims = ascii_lowercase[:len(shape.dims)]
        eq = f"{dims},{dims}->"
        eq = EinsumEquation(eq, shape, shape)
        contr = EinsumContraction(eq, op_shape=eq.shapes[0])
        return InnerGenerator(contr, 0, optimizer)

    def __call__[T: ArrayLike](
            self,
            guess: TrainBase[T],
            *operands: TrainBase[T],
            expr: bool = False) -> TrainBase[T]:
        trains_match(guess, *operands)
        guess = deepcopy(guess)
        gen_type = GeneratorCallableType.FULL

        inner_gens = [deepcopy(self._inner_gen) for _ in range(len(operands))]
        igens = [igen(guess, addend, expr=expr) for igen, addend in zip(inner_gens, operands)]
        lvec = local_vector(guess, self.decomposition)

        xp = namespace_of_trains(guess, *operands)
        data = xp.zeros(0)
        try:
            for _, lrange in self.strategy(guess.shape):
                lvec.send((lrange, data))
                arg = (lrange, gen_type)
                data = sum((igen.send(arg) for igen in igens[1:]),
                           start=igens[0].send(arg))
        finally:
            [igen.close() for igen in igens]
        return guess

    def calc_expressions(
            self,
            guess: TrainShape | TrainBase,
            operand: TrainShape | TrainBase) -> None:
        guess_shape = guess.shape if isinstance(guess, TrainBase) else guess
        op_shape = operand.shape if isinstance(operand, TrainBase) else operand
        self._inner_gen.calc_expressions(self.strategy, guess_shape, op_shape)
